"""Machinery for coordinates."""

from __future__ import annotations

import abc
from collections.abc import Sized
from functools import cache
from operator import gt, lt
from typing import Any, TypeVar

import numpy as np
import pandas as pd
from pydantic import (
    ValidationError,
    field_validator,
    model_serializer,
    model_validator,
)
from rich.text import Text
from typing_extensions import Self

import dascore as dc
from dascore.compat import array, is_array
from dascore.constants import dascore_styles
from dascore.exceptions import CoordError, ParameterError
from dascore.units import (
    Quantity,
    Unit,
    convert_units,
    get_factor_and_unit,
    get_quantity,
    get_quantity_str,
)
from dascore.utils.display import get_nice_text
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import (
    _get_nullish,
    _maybe_array_to_slice,
    _to_slice,
    _validate_sample_values,
    all_close,
    all_diffs_close_enough,
    cached_method,
    iterate,
    sanitize_range_param,
)
from dascore.utils.models import (
    ArrayLike,
    DascoreBaseModel,
    UnitQuantity,
)
from dascore.utils.time import dtype_time_like, is_datetime64, is_timedelta64, to_float

# Valid values for min/max
min_max_type = TypeVar("min_max_type")

step_type = TypeVar("step_type")


def ensure_consistent_dtype(value, name, dtype):
    """Ensure the values are consistent with dtype."""
    # For some reason all ints are getting converted to floats using default
    # pydantic type validation. This just fixes this manually.
    # TODO: See if this is needed in a few version after pydantic 2.1.1
    if pd.isnull(value):
        return value
    elif np.issubdtype(dtype, np.datetime64):
        if name == "step":
            value = dc.to_timedelta64(value)
        else:
            value = dc.to_datetime64(value)
    elif np.issubdtype(dtype, np.timedelta64):
        value = dc.to_timedelta64(value)
    # convert numpy numerics back to python
    elif np.issubdtype(dtype, np.floating):
        value = float(value) if value is not None else np.nan
    elif np.issubdtype(dtype, np.integer):
        value = int(value)
    return value


def _get_dtype(value, dtype):
    """Get the data type based on the first argument."""
    if dtype is not None and dtype != "":
        return str(dtype)
    value = type(value)
    return str(np.dtype(value))


class CoordSummary(DascoreBaseModel):
    """
    A summary for coordinates.

    Provides enough information for indexing coordinates and creating range
    coordinates.
    """

    dtype: str
    min: min_max_type
    max: min_max_type
    step: step_type | None = None
    units: UnitQuantity | None = None

    @model_serializer(when_used="json")
    def ser_model(self) -> dict[str, str]:
        """Serialize the model to json."""
        return {i: str(v) for i, v in self.model_dump().items()}

    @model_validator(mode="before")
    @classmethod
    def get_correct_dtype_cast_values(cls, data: Any) -> Any:
        """Ensure the correct dtype is provided and value conform to it."""
        if isinstance(data, dict):
            min_val = data["min"]
            dtype = _get_dtype(min_val, data.get("dtype"))
            data["dtype"] = str(dtype).split("[")[0]
            for name in ["min", "max", "step"]:
                val = data.get(name)
                data[name] = ensure_consistent_dtype(val, name, dtype)
        return data

    def to_coord(self) -> CoordRange:
        """Convert to coord range, if possible."""
        if pd.isnull(self.step):
            msg = "Cannot convert summary which is not evenly sampled to coord."
            raise CoordError(msg)
        step = self.step
        # this is a reverse coord
        if np.sign(step) == -1:
            start, stop = self.max, self.min + step
        else:
            start, stop = self.min, self.max + step
        return CoordRange(
            start=start,
            stop=stop,
            step=step,
            units=self.units,
        )


@cache
def _get_coord_filter_validators(dtype):
    """Get filter validators for a given input type."""

    def _is_sub_dtype(dtype1, dtype2):
        """Helper function to get sub dtypes."""
        # uncomment these if validators that arent numpy types are needed.
        # with suppress(TypeError):
        #     if issubclass(dtype1, dtype2):
        if np.issubdtype(dtype1, dtype2):
            return True
        return False

    # A list of dtype, func for validating/coercing single filter inputs.
    validators = (
        (pd.Timestamp, dc.to_datetime64),
        (np.datetime64, dc.to_datetime64),
        (pd.Timedelta, dc.to_timedelta64),
        (np.timedelta64, dc.to_timedelta64),
    )

    out = []
    for cls, func in validators:
        if _is_sub_dtype(dtype, cls):
            out.append(func)
    return tuple(out)


def get_compatible_values(val, dtype):
    """
    Get values compatible with dtype.

    This will essentially perform any type conversions needed to go from
    one dtype to another. It is useful for handling datetime conversions.

    Parameters
    ----------
    val
        The values to convert.
    dtype
        A numpy compatible datatype or string.
    """
    validators = _get_coord_filter_validators(dtype)
    for func in validators:
        if val is not None:
            val = func(val)
    return val


class BaseCoord(DascoreBaseModel, abc.ABC):
    """
    Coordinate interface.

    Coordinates are used to manage labels and indexing along a single
    data dimension.

    Coordinates should usually be created with
    [get_coords](`dascore.core.coords.get_coord`) rather than using the class
    directly.
    """

    units: UnitQuantity = None
    step: Any = None
    shape: tuple[int, ...] | None = None
    dtype: Any = None

    _rich_style = dascore_styles["default_coord"]
    _evenly_sampled = False
    _sorted = False
    _reverse_sorted = False
    _partial = False

    @model_validator(mode="before")
    @classmethod
    def check_time_units(cls, data: Any) -> Any:
        """Ensure time units are s if dattype is time-like."""
        if isinstance(data, dict):
            # This handles the coord range case.
            is_timey = False
            if start := data.get("start"):
                is_timey = is_timedelta64(start) or is_datetime64(start)
            elif (values := data.get("values")) is not None:
                is_timey = dtype_time_like(values)
            if is_timey and data.get("units") != (quant := get_quantity("s")):
                data["units"] = quant
        return data

    @field_validator("shape", mode="before")
    @classmethod
    def _validate_nullish_to_nan(cls, value):
        """Ensure shape is a tuple."""
        # This also allows shape to be an int.
        return tuple(iterate(value))

    @abc.abstractmethod
    def convert_units(self, unit) -> Self:
        """Convert from one unit to another. Set units if None are set."""

    def _get_value_index(self, coord_array, values_to_find):
        """Get the indices were values occur in array, account for duplicates."""
        # We check insertion order from both sides to catch duplicate values.
        inds_left = np.searchsorted(coord_array, values_to_find, side="left")
        inds_right = np.searchsorted(coord_array, values_to_find, side="right")
        if np.all(inds_right == (inds_left + 1)):  # Quick path for no duplicates.
            return inds_left
        # Each left right pair now needs to form a range so we include all
        # elements in between.
        ar = np.stack([inds_left, inds_right], axis=-1)
        inds = np.concatenate([np.arange(x[0], x[1]) for x in ar])
        return inds

    def _order_by_value_array(self, values_to_find):
        """Select values based on an array of values."""
        coord_array = self.values
        # First simply filter arg values to only include those in the index
        values_to_find = values_to_find[np.isin(values_to_find, coord_array)]
        # Handle fast cases for sorted and reverse sorted coords.
        if self.sorted:
            inds = self._get_value_index(coord_array, values_to_find)
            return self[inds], inds
        if self.reverse_sorted:
            # Need to_float here because datetime can't be multiplied by -1.
            inds = self._get_value_index(
                -to_float(coord_array), -to_float(values_to_find)
            )
            return self[inds], inds
        # Sort the array, then find insertion points, and map
        # back to pre-sorted indices.
        argsort = np.argsort(coord_array)
        sorted_coord_array = coord_array[argsort]
        sorted_inds = self._get_value_index(sorted_coord_array, values_to_find)
        inds = argsort[sorted_inds]
        return self[inds], inds

    def _order_by_sample_array(self, array):
        """Select based on index values."""
        if not np.issubdtype(array.dtype, np.integer):
            msg = (
                "Using an array input for select "
                "with samples requires integer dtype."
            )
            raise CoordError(msg)
        # Filter out bad indices
        array = array[np.abs(array) < len(self)]
        return self[array], array

    def _select_by_value_array(self, array):
        """Select values based on an array of values."""
        values = self.values
        # First simply filter arg values to only include those in the index
        valid_values = np.isin(values, array)
        return self[valid_values], valid_values

    def _select_by_sample_array(self, array):
        """Select based on index values."""
        if not np.issubdtype(array.dtype, np.integer):
            msg = "Using an array input for select with samples requires integer dtype."
            raise CoordError(msg)
        # Filter out bad indices
        assert self.ndim <= 1, "Select only works on 1D coords."
        inds = np.arange(len(self))
        valid_values = np.isin(inds, array)
        return self[valid_values], valid_values

    def _select_by_array(self, arg, samples=False, relative=False):
        """Select based on arg being an array."""
        if samples:
            return self._select_by_sample_array(arg)
        if np.issubdtype(getattr(arg, "dtype", None), np.bool_):
            return self[arg], arg
        arg = self._get_compatible_value(arg, relative=relative)
        return self._select_by_value_array(arg)

    def _select_by_samples(self, arg):
        """Select using samples."""
        _validate_sample_values(arg)
        reductions = _to_slice(arg)
        new = self[reductions]
        return new, reductions

    @abc.abstractmethod
    def select(
        self, arg, relative=False, samples=False
    ) -> tuple[Self, slice | ArrayLike]:
        """
        Returns an entity that can be used in a list for numpy indexing
        and selected coord.
        """

    def order(
        self, array, relative=False, samples=False
    ) -> tuple[Self, slice | ArrayLike]:
        """
        Order coordinate according to array values or samples.

        Parameters
        ----------
        array
            A numpy array of values in coordinate or (if samples)
            indices.
        relative
            If True, the values are relative to the start or end of coordinate.
        samples
            If True, the array is of dtype in and refers to samples in the
            coordinate.
        """
        array = np.atleast_1d(array)
        if samples:
            coord, inds = self._order_by_sample_array(array)
        else:
            array_compat = self._get_compatible_value(array, relative=relative)
            coord, inds = self._order_by_value_array(array_compat)
        return coord, _maybe_array_to_slice(inds, len(self))

    def align_to(
        self, other: BaseCoord
    ) -> tuple[BaseCoord, BaseCoord, slice | ArrayLike, slice | ArrayLike]:
        """
        Align the coordinate to another coordinate.

        This returns two new coordinates which share values as well indexer's
        needed to align corresponding arrays.

        Parameters
        ----------
        other
            The other coordinate.
        """

        def valid_non_coord(coord1, coord2):
            lens = {len(x) for x in [coord1, coord2]}
            # For compatibility one coord must have length 1 or
            # coords must be same length.
            if not (1 in lens or len(lens) == 1):
                msg = (
                    "Non coordinates must be the same length as coordinate "
                    "or length 1 for broadcasting to work."
                )
                raise CoordError(msg)

        if self == other:
            return self, other, slice(None), slice(None)
        assert self.ndim == 1, "can only align 1D arrays."
        if isinstance(self, CoordPartial) or isinstance(other, CoordPartial):
            valid_non_coord(self, other)
            return self, other, slice(None), slice(None)
        data1, data2 = self.data, other.data
        intersection = np.intersect1d(data1, data2)
        coord1, slice1 = self.order(intersection)
        coord2, slice2 = other.order(intersection)
        return coord1, coord2, slice1, slice2

    @abc.abstractmethod
    def __getitem__(self, item) -> Self:
        """Should implement slicing and return new instance."""

    @cached_method
    def __len__(self):
        """Total number of elements."""
        return self.shape[0]

    def __rich__(self):
        key_style = dascore_styles["keys"]
        base = Text("")
        base += Text(self.__class__.__name__, style=self._rich_style)
        base += Text("(")
        if not pd.isnull(self.min()):
            base += Text(" min: ", key_style)
            base += get_nice_text(self.min())
        if not pd.isnull(self.max()):
            base += Text(" max: ", key_style)
            base += get_nice_text(self.max())
        if not pd.isnull(self.step):
            base += Text(" step: ", key_style)
            base += get_nice_text(self.step)
        base += Text(" shape: ", key_style)
        base += get_nice_text(self.shape)
        base += Text(" dtype: ", key_style)
        base += get_nice_text(self.dtype)
        if self.units is not None:
            base += Text(" units: ", key_style)
            unit_str = get_quantity_str(self.units)
            base += get_nice_text(unit_str, style="units")
        base += Text(" )")
        return base

    def __str__(self):
        return str(self.__rich__())

    __repr__ = __str__

    def __array__(self, dtype=None, copy=False):
        """Numpy method for getting array data with `np.array(coord)`."""
        return self.data

    @cached_method
    def min(self):
        """Return min value."""
        return self._min()

    @cached_method
    def max(self):
        """Return max value."""
        return self._max()

    @property
    def unit_str(self) -> str:
        """Return a unit string."""
        return get_quantity_str(self.units)

    @abc.abstractmethod
    def _min(self):
        """Returns (or generates) the array data."""

    @abc.abstractmethod
    def _max(self):
        """Returns (or generates) the array data."""

    @property
    @cached_method
    def limits(self) -> tuple[Any, Any]:
        """Returns a numpy datatype."""
        return self.min(), self.max()

    @property
    @cached_method
    def ndim(self) -> int:
        """Return the number of dimensions in patch."""
        return len(self.shape)

    @property
    def size(self) -> int:
        """Return the size of the coordinate data."""
        return np.prod(self.shape)

    @property
    def evenly_sampled(self) -> tuple[int, ...]:
        """Returns True if the coord is evenly sampled."""
        return self._evenly_sampled

    @property
    def sorted(self) -> tuple[int, ...]:
        """Returns True if the coord in sorted."""
        return self._sorted

    @property
    def reverse_sorted(self) -> tuple[int, ...]:
        """Returns True if the coord in sorted in reverse order."""
        return self._reverse_sorted

    @property
    def degenerate(self) -> bool:
        """Return True of the coord is degenerate."""
        shape = self.shape
        return not len(shape) or np.prod(shape) == 0

    def set_units(self, units) -> Self:
        """Set new units on coordinates."""
        new = dict(self)
        new["units"] = units
        return self.__class__(**new)

    def simplify_units(self) -> Self:
        """Simplify the coordinate units."""
        _, unit = get_factor_and_unit(self.units, simplify=True)
        return self.convert_units(unit)

    def coord_range(self):
        """Return a scaler value for the coordinate (e.g., number of seconds)."""
        if not self.evenly_sampled:
            raise CoordError("coord_range has to be called on an evenly sampled data.")
        return self.max() - self.min() + self.step

    @abc.abstractmethod
    def sort(self, reverse=False) -> tuple[BaseCoord, slice | ArrayLike]:
        """Sort the contents of the coord. Return new coord and slice for sorting."""

    def snap(self) -> CoordRange:
        """
        Snap the coordinates to evenly sampled grid points.

        This will cause some loss of precision but often makes the data much
        easier to work with.
        """
        return self

    @abc.abstractmethod
    def update_limits(self, min=None, max=None, step=None, **kwargs) -> Self:
        """
        Update the limits or sampling of the coordinates.

        If start and stop are defined a new step is determined and returned.
        Next, the step size is updated changing only the end. Then the start
        is updated changing the start/end. Then the end is updated changing
        the start/end.

        Parameters
        ----------
        min
            The new start of the coordinate.
        max
            The new stop of the coordinate.
        step
            New step for the coordinate
        **kwargs
            Any other attributes which are used to create new coordinate.

        Notes
        -----
        For CoordRange stop will be max + step.
        """

    def update_data(
        self,
        data: ArrayLike | np.ndarray | None = None,
        values: ArrayLike | np.ndarray | None = None,
        **kwargs,
    ) -> Self:
        """
        Update the data of the coordinate.

        Parameters
        ----------
        data
            A new array to use.
        values
            Same as data, but deprecated. Here for compatibility reasons.
        """
        if data is None and values is None:
            return self
        data = values if data is None else data
        units = kwargs.get("units")
        return get_coord(data=data, units=units)

    def new(self, **kwargs):
        """Update coordinate."""
        info = self.model_dump(exclude_unset=True, exclude_defaults=True)
        # Need to only pass "values" rather than data to update
        if "data" in kwargs:
            kwargs["values"] = kwargs.pop("data")
        # Need to ensure new data is used in constructor, not old shape
        if "values" in kwargs:
            info.pop("shape", None)

        info.update(kwargs)
        return get_coord(**info)

    @property
    def data(self):
        """Return the internal data. Same as values attribute."""
        return self.values

    def _get_compatible_value(self, value, relative=False):
        """
        Return values that are compatible with dtype/units of coord.

        This is used, for example, to coerce values in select tuple
        so direct comparison with coord values is possible.
        """
        # strip units and v
        if hasattr(value, "units"):
            value = convert_units(value.magnitude, self.units, value.units)
        # if null or ... just return None
        if not is_array(value) and (pd.isnull(value) or value is Ellipsis):
            return None
        # special case for datetime and relative
        if relative:
            if np.issubdtype(self.dtype, np.datetime64):
                value = dc.to_timedelta64(value)
            value = self._get_relative_values(value)
        # apply validators. These can, eg, coerce to correct dtype.
        out = get_compatible_values(value, self.dtype)
        return out

    def _slice_degenerate(self, sliz):
        """
        Return bool indicating if the slice should yeild degenerate
        (empty array).
        """
        start, stop = sliz.start, sliz.stop
        # check if slice is between samples
        between = start is not None and start == stop
        # check if slice is outside of range
        bad_start = start is not None and (start < 0 or start >= len(self))
        bad_stop = stop is not None and (stop <= 0)
        return between or bad_start or bad_stop

    def get_slice_tuple(
        self,
        select: slice | None | type(Ellipsis) | tuple[Any, Any],
        relative=False,
    ) -> tuple[Any, Any]:
        """
        Get a tuple with (start, stop) and perform basic checks.

        Parameters
        ----------
        select
            An object for determining select range.
        """
        select_tuple = sanitize_range_param(select)
        p1, p2 = (
            self._get_compatible_value(x, relative=relative) for x in select_tuple
        )
        # reverse order if needed to ensure p1 < p2. This needs to be
        # after the compatible value conversion in case pre-converted
        # values are different types.
        if p1 is not None and p2 is not None and p2 < p1:
            p1, p2 = p2, p1
        return p1, p2

    def _get_relative_values(self, value):
        """Get relative values based on start (pos) or stop (neg)."""
        pos = np.sign(value).astype(np.int_) >= 0
        return self.min() + value if pos else self.max() + value

    def empty(self, axes=None) -> Self:
        """
        Empty out the coordinate.

        Parameters
        ----------
        axes
            The axis to empty, if None empty all.
        """
        if axes is None:
            new_shape = np.zeros(len(self.shape), dtype=np.int_)
        else:
            assert np.max(axes) <= (len(self) - 1)
            new_shape = np.asarray(self.shape)
            for ind in iterate(axes):
                new_shape[ind] = 0
        data = np.empty(tuple(new_shape), dtype=self.dtype)
        return get_coord(data=data)

    def index(self, indexer, axis: int | None = None) -> Self:
        """
        Index the coordinate and return new coordinate.

        Parameters
        ----------
        indexer
            Anything that can be used in numpy indexing.
        axis
            The axis along which to apply the indexer. If None,
            just apply indexer to numpy array.
        """
        if axis:
            ndims = len(self.shape)
            assert ndims >= (axis + 1)
            indexer = tuple(
                slice(None, None) if i != axis else indexer for i in range(ndims)
            )
        array = self.data[indexer]
        return get_coord(data=array, units=self.units)

    def get_attrs_dict(self, name):
        """Get attrs dict."""
        out = {f"{name}_min": self.min(), f"{name}_max": self.max()}
        if self.step:
            out[f"{name}_step"] = self.step
        if self.units:
            out[f"{name}_units"] = self.units
        return out

    def to_summary(self, dims=()) -> CoordSummary:
        """Get the summary info about the coord."""
        return CoordSummary(
            min=self.min(),
            max=self.max(),
            step=self.step,
            dtype=self.dtype,
            units=self.units,
        )

    def update(self, **kwargs):
        """Update parts of the coordinate."""
        out = self
        info = self.model_dump()
        update_fields = {
            i: v for i, v in kwargs.items() if not all_close(v, info.get(i))
        }
        units = update_fields.pop("units", None)
        _ = update_fields.pop("dtype", None)
        if update_fields:
            out = out.update_limits(**update_fields).update_data(**update_fields)
        if units is not None:
            out = out.convert_units(units)
        return out

    def get_sample_count(self, value, samples=False, enforce_lt_coord=False) -> int:
        """
        Return the number of samples represented by a value.

        This is calculated by dividing the value by dt and rounding up.
        Therefore, the output will always be greater or equal to 1.

        Parameters
        ----------
        value
            The value (supports units).
        samples
            If True, value is already in units of samples.
        """
        assert self.ndim == 1, "get sample count only works for 1D coords."
        if not self.evenly_sampled:
            msg = "Coordinate is not evenly sampled, can't get sample count."
            raise CoordError(msg)
        if samples:
            if not isinstance(value, int | np.integer):
                msg = "When samples==True values must be integers."
                raise ParameterError(msg)
            samples = int(value)
        else:
            compat_val = self._get_compatible_value(value, relative=True)
            duration = compat_val - self.min()
            samples = int(np.ceil(duration / self.step))
        if enforce_lt_coord and samples > len(self):
            msg = (
                f"value of {value} results in a window larger than coordinate "
                f"length of {len(self)}"
            )
            raise ParameterError(msg)
        return samples

    def get_next_index(
        self, value, samples=False, allow_out_of_bounds=False, relative=False
    ) -> int:
        """
        Get the index a value would have in a coordinate.

        This returns the "next" rather than the closest, index if the exact
        value is not contained by the index.

        Parameters
        ----------
        value
            The value which could be contained by the coordinate.
        samples
            If True, value refers to samples (ie an index) of coord.
        allow_out_of_bounds
            If True, allow the value to be out of bounds of the coordinate
            and just return an index referring to the end
            (len(coords) - 1) or beginning (0).
        relative
            If True, the provided values are relative to the start (if possitve)
            or end (if negative) of the coordinate.

        Examples
        --------
        >>> from dascore.core import get_coord
        >>> coord = get_coord(start=0, stop=10, step=1)
        >>> # Find the index for a value contained by the coordinate.
        >>> assert coord.get_next_index(1) == 1
        >>> # The next (not closest) index is return for value not in coord.
        >>> assert coord.get_next_index(2.000001) == 3
        """
        if not self.sorted:
            msg = f"Coords must be sorted to use get_next_index, {self} is not."
            raise CoordError(msg)
        input_array_like = isinstance(value, Sized)
        array = np.atleast_1d(value)

        # handle samples
        if samples:
            min_val, max_val = 0, len(self) - 1
            array = array.astype(np.int64)
            wrap_around = array < 0
            # account for negative indexing
            array[wrap_around] = array[wrap_around] + max_val + 1
        else:
            min_val, max_val = self.min(), self.max()
            array = self._get_compatible_value(array, relative=relative)
        # handle out of bounds cases
        is_gt, is_lt = array > max_val, array < min_val
        if not allow_out_of_bounds and np.any(is_gt | is_lt):
            msg = f"Value: {array} is out of bounds for {self}"
            raise ValueError(msg)
        # Fix max values
        array[is_gt] = max_val
        array[is_lt] = min_val
        # samples should already have the answer, just return
        if samples:
            return array if input_array_like else array[0]
        # otherwise get forward and backward inds
        forward_index = self._get_index(array, forward=True)
        back_index = self._get_index(array, forward=False)
        bad_for_index = pd.isnull(forward_index) | forward_index == -9999
        forward_index[bad_for_index] = back_index[bad_for_index]
        return forward_index if input_array_like else forward_index[0]

    def approx_equal(self: BaseCoord, other: BaseCoord) -> bool:
        """
        Return True if the coordinates are approximately equal.

        Parameters
        ----------
        other
            Another coordinate.
        """
        if self.shape != other.shape:
            return False
        non_coords = [self._partial, other._partial]
        if all(non_coords):
            return self == other
        if any(non_coords):
            return False
        return all_close(self.values, other.values)

    def change_length(self, length: int) -> Self:
        """
        Adjust the length of the coordinate by changing the end value.

        This is useful for floating point coordinates who frequently suffer
        from off by one errors.

        Note: Not all coordinates implement this method.

        Parameters
        ----------
        length
            The output length.
        """
        msg = f"Coordinate type {self.__class__} does not implement change_length"
        raise NotImplementedError(msg)


class CoordPartial(BaseCoord):
    """
    A coordinate which only contains partial information.
    """

    start: Any = np.nan
    stop: Any = np.nan
    step: Any = np.nan
    _rich_style = dascore_styles["coord_non"]
    _partial = True

    @field_validator("start", "stop", "step", mode="before")
    @classmethod
    def _validate_nullish_to_nan(cls, value, info):
        """Ensure nullish values are actually set as NaN"""
        if pd.isnull(value):
            return np.nan
        return value

    def __getitem__(self, item):
        # We init a temporary array just to get numpy to do the
        # indexing. There is probably a faster way but this is robust.
        dummy = np.empty(self.shape)[item]
        return self.__class__(shape=dummy.shape)

    def _max(self):
        """Dummy funct to do nothing but raise."""
        return self.stop

    def _min(self):
        return self.start

    def update(self, **kwargs):
        """No values to change so update can just call new."""
        return self.new(**kwargs)

    # Other operations that normally modify data do not in this case.
    update_limits = update
    set_units = update
    convert_units = update

    def sort(self, reverse=False):
        """Sort dummy array. Does nothing."""
        return self, slice(None, None)

    def __len__(self):
        return self.shape[0]

    @property
    def values(self):
        """Return the internal data. Same as values attribute."""
        null_val = np.asarray(_get_nullish(self.dtype))
        data = np.broadcast_to(null_val, self.shape)
        return data

    def _check_order_and_select(self, relative, samples):
        """Check that samples is True and relative false else raise msg."""
        if relative or not samples:
            msg = (
                "UnCoord does not support relative and samples must be True "
                "for both select and order methods."
            )
            raise CoordError(msg)

    def select(
        self, args, relative=False, samples=False
    ) -> tuple[BaseCoord, slice | ArrayLike]:
        """
        Select new values inside coord.

        For partial, samples==True or raise.
        """
        # Need to ensure relative is used OR the select has no effect.
        try:
            self._check_order_and_select(relative, samples)
        except CoordError as e:
            if not is_array(args):
                args = self.get_slice_tuple(args, relative=False)
                # Check if the select has no effect and return self or raise.
                if all(pd.isnull(x) for x in args):
                    return self, slice(None)
            raise e
        if is_array(args):
            return self._select_by_array(args, relative=relative, samples=samples)
        return self._select_by_samples(args)

    @compose_docstring(doc=BaseCoord.order.__doc__)
    def order(
        self, array, relative=False, samples=False
    ) -> tuple[Self, slice | ArrayLike]:
        """
        {doc}.
        """
        self._check_order_and_select(relative, samples)
        return super().order(array, relative=relative, samples=samples)

    @compose_docstring(doc=BaseCoord.change_length.__doc__)
    def change_length(self, length: int) -> Self:
        """
        {doc}
        """
        assert self.ndim == 1, "change_length only works on 1D coords."
        return get_coord(shape=(length,))

    def to_summary(self, dims=()) -> CoordSummary:
        """Get the summary info about the coord."""
        return CoordSummary(
            min=np.nan,
            max=np.nan,
            step=np.nan,
            dtype=self.dtype,
            units=None,
        )


class CoordRange(BaseCoord):
    """
    A coordinate represent a range of evenly sampled data.

    Parameters
    ----------
    start
        The starting value
    stop
        The ending value
    step
        The step between start and stop.

    Notes
    -----
    Like range and slice, CoordRange is exclusive of stop value.
    """

    start: Any = None
    stop: Any = None
    step: Any = None
    _evenly_sampled = True
    _rich_style = dascore_styles["coord_range"]

    @model_validator(mode="before")
    @classmethod
    def validate_start_stop_step_len(cls, values):
        """Coerce the needed values from the inputs."""
        req_values = ("start", "stop", "step", "shape")
        _attrs = [values.get(x, None) for x in req_values]
        valid_count = sum(not pd.isnull(x) for x in _attrs)
        if valid_count < 3:
            msg = (
                f"Three of {req_values} are required to create CoordRange. "
                f"You passed {values}"
            )
            raise CoordError(msg)
        # Now get start, stop, step from length, if provided.
        start, stop, step, shape = _attrs
        if not pd.isnull(shape):
            shape = tuple(iterate(shape))
            assert len(shape) == 1, "Coord range only works for 1D coords."
            length = shape[0]
            if pd.isnull(start):
                start = stop - step * length
            if pd.isnull(stop):
                stop = start + step * length
            if pd.isnull(step):
                step = (stop - start) / length
                # handle conversion to integer if other values are ints.
                if isinstance(start, int) and isinstance(stop, int):
                    step = int(step) if np.isclose(np.round(step), step) else step
        if step != 0:
            int_val = int(np.ceil(np.round((stop - start) / step, 1)))
            stop = start + step * int_val
        length = 1 if start == stop else int(np.round((stop - start) / step))
        shape = (length,)
        values.update(dict(start=start, stop=stop, shape=shape, step=step))
        # step should have the same sign as stop-start, see #321.
        diff = stop - start
        # note: we need to the to_float since np.sign(datetime64) returns a
        # datetime64 which includes precision, so even if the sign is the same
        # if the precision is different this validation fails.
        if not np.sign(to_float(values["step"])) == np.sign(to_float(diff)):
            msg = "Sign of step must match sign of stop - start"
            raise CoordError(msg)
        # Note: dtype was a property before but it messed up model
        # serialization.
        values["dtype"] = np.asarray(start + step).dtype
        return values

    def __getitem__(self, item):
        if isinstance(item, (int | np.integer)):
            if item >= len(self):
                raise IndexError(f"{item} exceeds coord length of {self}")
            return self.values[item]
        # handle ... as None
        if isinstance(item, slice):
            start = None if item.start is ... else item.start
            end = None if item.stop is ... else item.stop
            # Todo we can probably add more intelligent logic for slices.
            item = slice(start, end, item.step)
        out = self.values[item]
        return get_coord(data=out, units=self.units)

    @cached_method
    def __len__(self):
        return self.shape[0]

    def convert_units(self, units) -> Self:
        """Convert units, or set units if none exist."""
        # cant convert time units
        if dtype_time_like(self.dtype):
            return self
        out = dict(units=units)
        start = convert_units(self.start, to_units=units, from_units=self.units)
        stop = convert_units(self.stop, to_units=units, from_units=self.units)
        step = (stop - start) / len(self)
        out["start"], out["stop"], out["step"] = start, stop, step
        return self.__class__(**out)

    def select(
        self, args, relative=False, samples=False
    ) -> tuple[BaseCoord, slice | ArrayLike]:
        """
        Apply select, return selected coords and index to apply to array.

        Can return a CoordDegenerate if selection is outside of range.
        """
        if is_array(args):
            return self._select_by_array(args, relative=relative, samples=samples)
        elif samples:
            return self._select_by_samples(args)
        args = self.get_slice_tuple(args, relative=relative)
        start = self._get_index(args[0], forward=self.sorted)
        stop = self._get_index(args[1], forward=self.reverse_sorted)
        if self.reverse_sorted:
            start, stop = stop, start
        # we add 1 to stop in slice since its upper limit is exclusive
        start = None if start == 0 else start
        data = slice(start, (stop + 1) if stop is not None else stop)
        if self._slice_degenerate(data):
            return self.empty(), slice(0, 0)
        new_start = self[start] if start is not None else self.start
        new_end = self[stop] + self.step if stop is not None else self.stop
        new_coords = self.new(start=new_start, stop=new_end)
        return new_coords, data

    def sort(self, reverse=False) -> tuple[BaseCoord, slice | ArrayLike]:
        """Sort the contents of the coord. Return new coord and slice for sorting."""
        #
        forward_forward = not reverse and self.sorted
        reverse_reverse = reverse and self.reverse_sorted
        if forward_forward or reverse_reverse:
            return self, slice(None)
        new_step = -self.step
        if reverse:  # reversing a forward sorted Coordrange
            new_start, new_stop = self.max(), self.min() + new_step
        else:  # order a reverse sorted one
            new_start, new_stop = self.min(), self.max() + new_step
        out = self.new(start=new_start, stop=new_stop, step=new_step)
        return out, slice(None, None, -1)

    def _get_index(self, value, forward=True):
        """Get the index corresponding to a value."""
        if (value := self._get_compatible_value(value)) is None:
            return value
        input_is_array = isinstance(value, Sized)
        array = np.atleast_1d(value)
        func = np.ceil if forward else np.floor
        start, step = self.start, self.step
        # Due to float weirdness we need a little bit of a fudge factor here.
        fraction = func(np.round((array - start) / step, decimals=10))
        out = fraction.astype(np.int64)
        lt_forward = (out < 0) & forward
        gt_back = (out >= len(self)) & (not forward)
        bad_values = lt_forward | gt_back
        if not input_is_array and np.any(bad_values):
            return None
        return out if input_is_array else int(out[0])

    @compose_docstring(doc=BaseCoord.update_limits.__doc__)
    def update_limits(self, min=None, max=None, step=None, **kwargs) -> Self:
        """{doc}."""
        if all(x is not None for x in [min, max, step]):
            msg = "At most two parameters can be specified in update_limits."
            raise ValueError(msg)
        # first case, we need to determine new dt.
        if min is not None and max is not None:
            new_step = (max - min) / len(self)
            return get_coord(start=min, stop=max, step=new_step, units=self.units)
        # For other combinations we just apply adjustments sequentially
        # after ensuring that the types are compatible.
        out = self
        if step is not None:
            step = get_compatible_values(step, type(self.step))
            new_stop = out.start + step * len(out)
            out = out.new(stop=new_stop, step=step)
        if min is not None:
            min = get_compatible_values(min, self.dtype)
            diff = min - out.start
            new_stop = out.stop + diff
            out = out.new(start=min, stop=new_stop)
        if max is not None:
            max = get_compatible_values(max, self.dtype)
            translation = (max + out.step) - out.stop
            new_start = self.start + translation
            # we add step so the new range is inclusive of stop.
            out = out.new(start=new_start, stop=max + out.step)
        return out.new(**kwargs)

    @property
    @cached_method
    def values(self) -> ArrayLike:
        """Return the values of the coordinate as an array."""
        if len(self) == 1:
            return np.asarray([self.start])
        # note: linspace works better for floats that might have slightly
        # uneven spacing. It ensures the length of the output array is robust
        # to small deviations in spacing. However, this doesnt work for datetimes.
        if is_datetime64(self.start) or is_timedelta64(self.start):
            out = np.arange(self.start, self.stop, self.step)
        else:
            out = np.linspace(
                self.start, self.stop - self.step, num=len(self), dtype=self.dtype
            )
        # again, due to round-off error the array can one element longer than
        # anticipated. The slice here just ensures shape and len match.
        return array(out[: len(self)])

    def _min(self):
        """Return min value."""
        return np.min([self.start, self.stop - self.step])

    def _max(self):
        """Return max value in range."""
        # like range, coord range is exclusive of final value.
        # the min/max are needed for reverse sorted coord.
        return np.max([self.stop - self.step, self.start])

    @compose_docstring(doc=BaseCoord.change_length.__doc__)
    def change_length(self, length: int) -> Self:
        """
        {doc}
        """
        assert self.ndim == 1, "Can only change length for 1D coords."
        if (current := len(self)) == length:
            return self
        diff = length - current
        stop, step = self.stop, self.step
        out = self.update(stop=stop + step * diff)
        assert len(out) == length
        return out

    @property
    def sorted(self) -> bool:
        """Returns true if sorted in ascending order."""
        return self.step >= 0

    @property
    def reverse_sorted(self) -> bool:
        """Returns true if sorted in ascending order."""
        return self.step < 0


class CoordArray(BaseCoord):
    """
    A coordinate with arbitrary values in an array.

    Can handle any number of dimensions.
    """

    values: ArrayLike
    _rich_style = dascore_styles["coord_array"]

    @model_validator(mode="before")
    @classmethod
    def validate_start_stop_step_len(cls, values):
        """Coerce the needed values from the inputs."""
        values["dtype"] = values["values"].dtype
        values["shape"] = values["values"].shape
        return values

    def convert_units(self, units) -> Self:
        """Convert units, or set units if none exist."""
        is_time = np.issubdtype(self.dtype, np.datetime64)
        is_time_delta = np.issubdtype(self.dtype, np.timedelta64)
        if self.units is None or is_time or is_time_delta:
            return self.set_units(units)
        values = convert_units(self.values, units, self.units)
        return self.new(units=units, values=values)

    def select(
        self, args, relative=False, samples=False
    ) -> tuple[Self, slice | ArrayLike]:
        """Apply select, return selected coords and index for selecting data."""
        if is_array(args):
            return self._select_by_array(args, relative=relative, samples=samples)
        elif samples:
            return self._select_by_samples(args)

        args = self.get_slice_tuple(args, relative=relative)
        values = self.values
        out = np.ones_like(values, dtype=np.bool_)
        val1 = self._get_compatible_value(args[0])
        val2 = self._get_compatible_value(args[1])
        if val1 is not None:
            out = out & (values >= val1)
        if val2 is not None:
            out = out & (values <= val2)
        if not np.any(out):
            return self.empty(), out
        if np.all(out):
            return self, slice(None, None)
        # Convert boolean to int indexes because these are supported for
        # indexing pytables arrays but booleans are not.
        if len(self.shape) == 1:
            out = np.arange(len(out))[out]
        return self.new(values=values[out]), out

    def sort(self, reverse=False) -> tuple[BaseCoord, slice | ArrayLike]:
        """Sort the coord to be monotonic (maybe range)."""
        argsort: ArrayLike = np.argsort(self.values)[:: -1 if reverse else 1]
        arg_dict = self.model_dump()
        arg_dict["values"] = self.values[argsort]
        new = get_coord(**arg_dict)
        return new, argsort

    def snap(self):
        """
        Snap the coordinates to evenly sampled grid points.

        This will cause some loss of precision but often makes the coordinate
        much easier to work with. The min/max of the coordinate will remain
        unchanged.
        """
        values = self.values
        min_v, max_v = np.min(values), np.max(values)
        if len(self) == 1:
            # time deltas need to be generated for dt case, hence the subtract
            _zero = self._get_compatible_value(0)
            step = self._get_compatible_value(1) - _zero
            # we just use a step of 1 in case of len 1 coord.
        else:
            dur = max_v - min_v
            is_dt = is_timedelta64(dur)
            # hack to handle dts int division.
            if is_dt:
                _step = float(dur.astype(np.int64)) / (len(self) - 1)
                step = np.timedelta64(int(np.round(_step)), "ns")
            else:
                step = dur / (len(self) - 1)
            assert step > 0
        if self.reverse_sorted:
            step = -step
            start, stop = max_v, min_v + step
        else:
            start, stop = min_v, max_v + step
        # Get potential output, ensure it is the same length as original.
        out = CoordRange(start=start, stop=stop, step=step, units=self.units)
        return out.change_length(len(self))

    @compose_docstring(doc=BaseCoord.update_limits.__doc__)
    def update_limits(self, min=None, max=None, step=None, **kwargs) -> Self:
        """{doc}."""
        if sum(x is not None for x in [min, max, step]) > 1:
            msg = "At most one parameter can be specified in update_limits."
            raise ValueError(msg)
        out = self
        if not pd.isnull(step) and len(self):
            out = self.snap().update_limits(step=step)
        elif min is not None:
            diff = min - self.min()
            vals = self.values + diff
            out = get_coord(data=vals, units=self.units)
        elif max is not None:
            diff = max - self.max()
            vals = self.values + diff
            out = get_coord(data=vals, units=self.units)
        return out.new(**kwargs)

    def __getitem__(self, item) -> Self:
        out = self.values[item]
        if not np.ndim(out):
            return out
        return self.__class__(values=out, units=self.units)

    def _min(self):
        """Return min value."""
        return np.nanmin(self.values)

    def _max(self):
        """Return max value in range."""
        return np.nanmax(self.values)


class CoordMonotonicArray(CoordArray):
    """A coordinate with strictly increasing or decreasing values."""

    values: ArrayLike
    _rich_style = dascore_styles["coord_monotonic"]
    _sorted = True

    def select(
        self, args, relative=False, samples=False
    ) -> tuple[Self, slice | ArrayLike]:
        """Apply select, return selected coords and index for selecting data."""
        if is_array(args):
            return self._select_by_array(args, relative=relative, samples=samples)
        elif samples:
            return self._select_by_samples(args)

        v1, v2 = self.get_slice_tuple(args, relative=relative)
        # reverse order if reverse monotonic. This is done so when we mult
        # by -1 in _get_index the inverted range is used.
        if self.reverse_sorted:
            v1, v2 = v2, v1
        start = self._get_index(v1, forward=False)
        new_start = start if start is not None and start > 0 else None
        stop = self._get_index(v2, forward=True)
        new_stop = stop if stop is not None and stop < len(self) else None
        # We need to add 1 to end so 1 sample get selected if start == stop
        if new_stop is not None:
            if self.values[new_stop] == v2:
                new_stop = new_stop + 1
        out = slice(new_start, new_stop)
        if self._slice_degenerate(out):
            return self.empty(), slice(0, 0)
        return self.new(values=self.values[out]), out

    def _get_index(self, value, forward=True):
        """
        Get the index corresponding to a value.

        Forward indicates if this is the max (left) value.
        """
        if (new_value := self._get_compatible_value(value)) is None:
            return new_value
        values = np.atleast_1d(self.values)
        # since search sorted only works on ascending monotonic arrays we
        # negative descending arrays to get the same effect.
        if self.reverse_sorted:
            values = to_float(values) * -1
            new_value = to_float(new_value) * -1
        # side = "right" if forward else "left"
        # out = np.atleast_1d(np.searchsorted(values, new_value, side=side))
        # Search values. Ensure the returned index is in bounds (eg values GT
        # coord max should still have a range in coords.
        new_value = np.atleast_1d(new_value)
        right = np.searchsorted(values, new_value, side="right")
        # right_ok = (right < len(self)) & (right < 0)
        left = np.searchsorted(values, new_value, side="left")
        left_ok = (left < len(self)) & (left > 0)
        eq = left_ok & (values.take(left, mode="clip") == new_value)
        out = right if forward else left
        # where equal it should also be left values. This makes the function
        # behavior consistent with BaseCoord._get_index.
        if not self.reverse_sorted:
            out[eq] = left[eq]
        return out if is_array(value) else int(out[0])

    def _step_meets_requirement(self, op):
        """Return True is any data increment meets the comp. requirement."""
        vals = self.values
        # we must iterate because first two elements might be equal.
        # but this wont iterate the whole array; just until sort order is found
        for ind in range(1, len(self)):
            if op(vals[ind], vals[ind - 1]):
                return True
        # we consider single valued arrays sorted, but not reverse sorted.
        if len(vals) == 1 and op is gt:
            return True
        return False

    @property
    @cached_method
    def sorted(self):
        """Determine is coord array is sorted in ascending order."""
        return self._step_meets_requirement(gt)

    @property
    @cached_method
    def reverse_sorted(self):
        """Determine is coord array is sorted in descending order."""
        return self._step_meets_requirement(lt)


def get_coord(
    *,
    data: ArrayLike | None | np.ndarray | BaseCoord = None,
    values: ArrayLike | None | np.ndarray = None,
    start=None,
    min=None,
    stop=None,
    max=None,
    step=None,
    units: None | Unit | Quantity | str = None,
    shape: None | int | tuple[int, ...] = None,
    dtype: str | np.dtype = None,
) -> BaseCoord:
    """
    Return a coordinate from provided inputs.

    This function figures out which kind of Coordinate should be returned
    for provided inputs.

    Parameters
    ----------
    data
        An array indicating the values or an integer to specify the length
        of a partial coordinate.
    values
        Deprecated, use data instead.
    start
        The start value of the array, inclusive.
    min
        The minimum value, same as start.
    stop
        The stopping value of an array, exclusive.
    step
        The sampling spacing of an array.
    units
        Indication of units.
    shape
        If an int or tuple, the output should be a partial coord of with
        this shape. Otherwise, leave unset.
    dtype
        Data type for coord. Often can be inferred from other arguments.

    Notes
    -----
    The following combinations of input parameters are typical:
        (start, stop, step)
        (values)
        (values, step) - useful for length 1 arrays.

    Examples
    --------
    >>> import numpy as np
    >>> from dascore.core import get_coord
    >>>
    >>> # Create a coordinate from a start, stop, and range value.
    >>> range_coord = get_coord(start=1, stop=12, step=1)
    >>>
    >>> # Create an identical coordinate from an array.
    >>> array_coord = get_coord(data=np.arange(1, 12, 1))
    >>> # This array coord should return an identical coordinate
    >>> assert range_coord == array_coord
    >>>
    >>> # Coordinate from an array that is sorted, but not evenly sampled
    >>> array = np.sort(np.random.rand(20))
    >>> array_coord2 = get_coord(data=array)
    >>>
    >>> # Coordinate from random array
    >>> array = np.random.rand(20)
    >>> array_coord3 = get_coord(data=array)
    >>>
    >>> # Create a partial coordinate of a given shape
    >>> partial_coord = get_coord(shape=(10,))
    """

    def _check_data_compatibility(data, start, stop, step):
        """Ensure input combinations are valid."""
        if data is None:
            if any([start is None, stop is None, step is None]):
                msg = "When data is not defined, start, stop, and step must be."
                raise CoordError(msg)

    def _get_new_max(data, min, step):
        """Get the new length to use."""
        # for int based data types we need to modify the end time
        # otherwise this will just go nuts
        dtype = getattr(min, "dtype", None)
        if dtype_time_like(dtype) or np.issubdtype(dtype, np.integer):
            max = min + (len(data) - 1) * step
        else:
            max = data[-1]
        return max

    def _get_shape(shape):
        """Return proper shape tuple or None."""
        if shape is None or isinstance(shape, Sized):
            return shape
        return (shape,)

    def _get_array(data, values):
        """Get the array from either data or values."""
        # ensure data and values are not used
        if data is not None and values is not None:
            msg = "Cannot specify both data and values. Use only data."
            raise CoordError(msg)
        elif values is not None:
            data = values
        return data

    def _maybe_get_start_stop_step(data):
        """Get start, stop, step, is_monotonic."""
        data = np.asarray(data)
        view2 = data[1:]
        view1 = data[:-1]
        is_monotonic = np.all(view1 > view2) or np.all(view2 > view1)
        # the array cannot be evenly sampled if it isn't monotonic
        if is_monotonic:
            diffs = view2 - view1
            unique_diff = np.unique(diffs)
            if len(unique_diff) == 1 or all_diffs_close_enough(unique_diff):
                _min = data[0]
                # this is a poor man's median that preserves dtype
                sorted_diffs = np.sort(diffs)
                _step = sorted_diffs[len(sorted_diffs) // 2]
                _max = _get_new_max(data, _min, _step)
                return _min, _max + _step, _step, is_monotonic
        return None, None, None, is_monotonic

    data = _get_array(data, values)
    shape = _get_shape(shape)
    if data is None and shape is not None:
        attrs = dict(
            shape=shape, start=start, stop=stop, step=step, units=units, dtype=dtype
        )
        try:  # This could be a normal RangeCoord
            return CoordRange(**attrs)
        except (ValidationError, CoordError):  # If not it's a partial
            return CoordPartial(**attrs)

    # maybe convert min/max to start stop.
    if start is None and min is not None:
        start = min
    if stop is None and max is not None:
        stop = max
    _check_data_compatibility(data, start, stop, step)
    # data array was passed; see if it is monotonic/evenly sampled
    if data is not None:
        # Handle attached units.
        if isinstance(data, dc.units.Quantity):
            data, maybe_units = data.magnitude, data.units
            units = units if units is not None else maybe_units
        if isinstance(data, (int | np.integer)):
            shape = _get_shape(data)
            attrs = dict(
                shape=shape, start=start, stop=stop, step=step, units=units, dtype=dtype
            )
            return CoordPartial(**attrs)
        if isinstance(data, BaseCoord):  # just return coordinate
            return data
        if not isinstance(data, np.ndarray):
            data = np.atleast_1d(data)
        if np.size(data) == 0:
            dtype = dtype or data.dtype
            return CoordPartial(shape=data.shape, units=units, step=step, dtype=dtype)
        # special case of len 1 array either get range, if step specified
        # or sorted monotonic array if not.
        elif len(data) == 1:
            if not pd.isnull(step):
                val = data[0]
                return CoordRange(start=val, stop=val + step, step=step, units=units)
            return CoordMonotonicArray(values=data, units=units)
        start, stop, step, monotonic = _maybe_get_start_stop_step(data)
        if start is not None:
            out = CoordRange(start=start, stop=stop, step=step, units=units)
            # The change_length call helps with float off by one issues.
            return out.change_length(len(data))
        elif monotonic:
            return CoordMonotonicArray(values=data, units=units)
        elif np.all(pd.isnull(data)):
            return CoordPartial(
                shape=data.shape,
                units=units,
                start=start,
                stop=stop,
                step=step,
                dtype=dtype,
            )
        return CoordArray(values=data, units=units)
    else:
        return CoordRange(start=start, stop=stop, step=step, units=units)



"""
Core routines and functionality for processing distributed fiber data.
"""
from __future__ import annotations
from .patch import Patch  # noqa
from .coords import CoordSummary, get_coord
from .coordmanager import get_coord_manager, CoordManager



"""Tests for taper processing function."""

from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import ParameterError
from dascore.proc.taper import TAPER_FUNCTIONS, taper
from dascore.units import m
from dascore.utils.misc import broadcast_for_index

gen = np.random.default_rng(32)


@pytest.fixture(scope="session")
def patch_ones(random_patch):
    """Return a patch filled with ones."""
    patch = random_patch.new(data=np.ones_like(random_patch.data))
    return patch


@pytest.fixture(scope="session", params=sorted(TAPER_FUNCTIONS))
def time_tapered_patch(request, patch_ones):
    """Return a tapered trace."""
    # first get a patch with all ones for easy testing
    patch = patch_ones.update(data=np.ones_like(patch_ones.data))
    out = taper(patch, time=0.05, window_type=request.param)
    return out


def _get_start_end_indices(patch, dim):
    """Helper function to get indices for slicing start/end of data."""
    axis = patch.dims.index(dim)
    n_dims = len(patch.dims)
    inds_start = broadcast_for_index(n_dims, axis, 0)
    inds_end = broadcast_for_index(n_dims, axis, -1)
    return inds_start, inds_end


class TestTaperBasics:
    """Ensure each taper runs."""

    def test_each_taper(self, time_tapered_patch, random_patch):
        """Ensure each taper type runs."""
        assert isinstance(time_tapered_patch, dc.Patch)
        assert time_tapered_patch.shape == random_patch.shape

    def test_time_dt_unchanged(self, time_tapered_patch, random_patch):
        """Ensure each taper type runs."""
        attrs1, attrs2 = random_patch.attrs, time_tapered_patch.attrs
        assert attrs1.time_units == attrs2.time_units
        assert attrs1.time_step == attrs2.time_step

    def test_ends_near_zero(self, time_tapered_patch):
        """Ensure the ends of the patch are near zero."""
        patch = time_tapered_patch
        data = patch.data
        inds_start, inds_end = _get_start_end_indices(patch, "time")
        assert np.all(data[inds_start] < 0.09)
        assert np.all(data[inds_end] < 0.09)

    def test_taper_start_only(self, patch_ones):
        """Ensure tapering only start works."""
        patch = patch_ones.taper(time=(0.05, None))
        data = patch.data
        inds_start, inds_end = _get_start_end_indices(patch, "time")
        assert np.all(data[inds_start] < 0.09)
        assert np.allclose(data[inds_end], 1)

    def test_taper_end_only(self, patch_ones):
        """Ensure tapering only the end works."""
        patch = patch_ones.taper(time=(None, 0.05))
        data = patch.data
        inds_start, inds_end = _get_start_end_indices(patch, "time")
        assert np.all(data[inds_end] < 0.09)
        assert np.allclose(data[inds_start], 1)

    def test_overlapping_windows_raises(self, patch_ones):
        """Overlapping tapers should raise."""
        with pytest.raises(ParameterError, match="cannot overlap"):
            patch_ones.taper(time=(0.51, 0.51))

    def test_too_big_raises(self, patch_ones):
        """Single tapers too large should raise."""
        with pytest.raises(ParameterError, match="taper lengths exceed"):
            patch_ones.taper(time=(None, 1.01))
        with pytest.raises(ParameterError, match="taper lengths exceed"):
            patch_ones.taper(time=(1.01, None))

    def test_doc_example(self, random_patch):
        """Tests the doc example case."""
        patch = random_patch
        patch_taper1 = patch.taper(time=0.05, window_type="hann")
        patch_taper2 = patch.taper(distance=(0.10, None), window_type="triang")
        assert patch_taper1.shape == patch_taper2.shape

    def test_bad_taper_str_raises(self, random_patch):
        """Test that an invalid taper function raises nice message."""
        with pytest.raises(ParameterError, match="not a known window"):
            random_patch.taper(time=(0.1, 0.1), window_type="windowsXP")
            # the only good one...

    def test_taper_with_units(self, patch_ones):
        """Ensure taper words with units specified."""
        value = 15 * m
        patch = patch_ones.taper(distance=value)
        data_new = patch.data
        data_old = patch_ones.data
        dim_len = patch.dims.index("distance")
        mid_dim = dim_len // 2

        assert not np.allclose(data_new, data_old)

        assert np.allclose(
            data_new[mid_dim - 10 : mid_dim + 10], data_old[mid_dim - 10 : mid_dim + 10]
        )

    def test_timedelta_taper(self, random_patch):
        """Test that a timedelta works for the taper argument. See #379."""
        time1 = dc.to_timedelta64(2)
        time2 = 2 * dc.get_quantity("seconds")
        patch1 = random_patch.taper(time=time1)
        patch2 = random_patch.taper(time=time2)
        assert patch1 == patch2


class TestTaperRange:
    """Test for tapering a range of values."""

    @pytest.fixture(scope="class")
    def patch_sorted_time(self, random_patch):
        """Return a patch with sorted but not evenly spaced time dim."""
        times = gen.random(len(random_patch.get_coord("time")))
        new_times = dc.to_datetime64(np.sort(times))
        return random_patch.update_coords(time=new_times)

    def test_dims(self, patch_ones):
        """Ensure both dimensions work."""
        for ax, dim in enumerate(patch_ones.dims):
            coord = patch_ones.get_coord(dim)
            clen = len(coord)
            ind1 = int(clen / 3)
            ind2 = int(2 * clen / 3)
            val1, val2 = coord[ind1], coord[ind2]
            out = patch_ones.taper_range(**{dim: (val1, val2)})
            data = out.data
            # Ensure the center of the envelope are untouched
            tapered_inds = broadcast_for_index(data.ndim, ax, slice(ind1, ind2))
            tapered = data[tapered_inds]
            assert np.allclose(tapered, 1)
            # Ensure the edges are near 0
            start_vals = data[broadcast_for_index(data.ndim, ax, 0)]
            assert np.allclose(start_vals, 0)
            end_vals = data[broadcast_for_index(data.ndim, ax, -1)]
            assert np.allclose(end_vals, 0)

    def test_invert(self, patch_ones):
        """Ensure inverting results in 0s near the peaks of taper."""
        dim, ax = "time", patch_ones.dims.index("time")
        coord = patch_ones.get_coord(dim)
        clen = len(coord)
        ind1 = int(clen / 3)
        ind2 = int(2 * clen / 3)
        val1, val2 = coord[ind1], coord[ind2]
        out = patch_ones.taper_range(**{dim: (val1, val2)}, invert=True)
        data = out.data
        # Ensure the center of the envelope are untouched
        tapered_inds = broadcast_for_index(data.ndim, ax, slice(ind1, ind2))
        tapered = data[tapered_inds]
        assert np.allclose(tapered, 0)
        # Ensure the edges are near 0
        start_vals = data[broadcast_for_index(data.ndim, ax, 0)]
        assert np.allclose(start_vals, 1)
        end_vals = data[broadcast_for_index(data.ndim, ax, -1)]
        assert np.allclose(end_vals, 1)

    def test_four_value_taper(self, patch_ones):
        """Ensure the range works with 4 values specifying taper limits."""
        dim = "time"
        coord = patch_ones.get_coord(dim)
        clen = len(coord)
        inds = int(clen / 6), int(2 * clen / 6), int(4 * clen / 6), int(5 * clen / 6)
        vals = [coord[x] for x in inds]
        out = patch_ones.taper_range(**{dim: vals}, invert=False)

        in_taper = out.select(time=(vals[1], vals[2]))
        assert np.allclose(in_taper.data, 1, rtol=1e-4)

        right_of_taper = out.select(time=(vals[3], ...))
        assert np.allclose(right_of_taper.data, 0)
        left_of_taper = out.select(time=(..., vals[0]))
        assert np.allclose(left_of_taper.data, 0)

    def test_samples_and_relative(self, random_patch):
        """Ensure samples and relative work with coord."""
        coord = random_patch.get_coord("time")
        values = coord.values
        ind = 200
        # normal style taper
        p1 = random_patch.taper_range(time=(values[ind], values[-ind]))
        # relative taper
        time_range = (values[ind] - values[0], values[-ind] - values[-1])
        p2 = random_patch.taper_range(time=time_range, relative=True)
        # samples taper
        p3 = random_patch.taper_range(time=(ind, -ind), samples=True)
        assert p1.equals(p2) and p2.equals(p3)

    def test_poorly_shaped_sequence_raises(self, random_patch):
        """Ensure sequences of wrong shape raise."""
        with pytest.raises(ParameterError, match="sequence is required"):
            random_patch.taper_range(time=1)
        with pytest.raises(ParameterError, match="sequence is required"):
            random_patch.taper_range(time=[1])
        with pytest.raises(ParameterError, match="sequence is required"):
            random_patch.taper_range(time=[1] * 10)

    def test_bad_use_of_none(self, random_patch):
        """Ensure bad use of None raises."""
        with pytest.raises(ParameterError, match="Cannot use ... or None"):
            random_patch.taper_range(time=(1, None), relative=True)

    def test_use_none(self, random_patch):
        """Ensure use of None specifies proper limits."""
        coord = random_patch.get_coord("time")
        time_span = coord.max() - coord.min()
        out1 = random_patch.taper_range(time=(None, 1, 4, None), relative=True)
        out2 = random_patch.taper_range(time=(0, 1, 4, time_span), relative=True)
        assert out1.equals(out2)

    def test_multiple_taper_values(self, patch_ones):
        """Ensure multiple values in taper are possible."""
        taper_range = ((25, 50, 100, 125), (150, 175, 200, 225))
        out = patch_ones.taper_range(distance=taper_range)

        # Check a few points inside and outside of the taper bounds
        assert np.allclose(out.select(distance=(..., 22)).data, 0)
        assert np.allclose(out.select(distance=(55, 90)).data, 1)
        assert np.allclose(out.select(distance=(126, 149)).data, 0)
        assert np.allclose(out.select(distance=(225, ...)).data, 0)

    def test_uneven_coords(self, patch_sorted_time):
        """Ensure uneven coords also work for tapering."""
        out = patch_sorted_time.taper_range(time=(0.4, 0.6), relative=True)
        assert isinstance(out, dc.Patch)

    def test_non_cosine_window(self, patch_ones):
        """Ensure other window funcs also work."""
        taper_range = ((25, 50, 100, 125), (150, 175, 200, 225))
        out = patch_ones.taper_range(distance=taper_range, window_type="ramp")
        # Check a few points inside and outside of the taper bounds
        assert np.allclose(out.select(distance=(..., 22)).data, 0)
        assert np.allclose(out.select(distance=(55, 90)).data, 1)
        assert np.allclose(out.select(distance=(126, 149)).data, 0)
        assert np.allclose(out.select(distance=(225, ...)).data, 0)



"""A spool for working with a single file."""

from __future__ import annotations

import copy
from pathlib import Path

from rich.text import Text
from typing_extensions import Self

import dascore as dc
from dascore.constants import PROGRESS_LEVELS, SpoolType
from dascore.core.spool import BaseSpool, DataFrameSpool
from dascore.io.core import FiberIO
from dascore.utils.docs import compose_docstring


class FileSpool(DataFrameSpool):
    """
    A spool for a single file.

    Parameters
    ----------
    path
        The path to the file.
    file_format
        The format name, optional.
    file_version
        The version string of the format, optional.

    Notes
    -----
    Some file formats support storing multiple patches, this is most useful
    for those formats, but should work on all dascore supported formats.
    """

    def __init__(
        self,
        path: str | Path,
        file_format: str | None = None,
        file_version: str | None = None,
    ):
        super().__init__()
        # Init file spool from another file spool
        if isinstance(path, self.__class__):
            self.__dict__.update(copy.deepcopy(path.__dict__))
            return
        self._path = Path(path)
        if not self._path.exists() or self._path.is_dir():
            msg = f"{path} does not exist or is a directory"
            raise FileNotFoundError(msg)

        _format, _version = dc.get_format(path, file_format, file_version)
        source_df = dc.scan_to_df(path, file_format=_format, file_version=_version)
        dfs = self._get_dummy_dataframes(source_df)
        self._df, self._source_df, self._instruction_df = dfs
        self._file_format = _format
        self._file_version = _version

    def __rich__(self):
        """Augment rich string with path."""
        base = super().__rich__()
        out = base + Text(f" Path: {self._path}")
        return out

    def _load_patch(self, kwargs) -> Self:
        """Given a row from the managed dataframe, return a patch."""
        return dc.read(**kwargs)[0]

    @compose_docstring(doc=BaseSpool.update.__doc__)
    def update(self: SpoolType, progress: PROGRESS_LEVELS = "standard") -> Self:
        """
        {doc}.

        Note: If the file format supports indexing (e.g. DASDAE) this will
        trigger an indexing of the file.
        """
        formater = FiberIO.manager.get_fiberio(
            format=self._file_format, version=self._file_version
        )
        getattr(formater, "index", lambda x: None)(self._path)
        return self



"""IO module for reading SEGY file format support."""

from __future__ import annotations

import dascore as dc
from dascore.io.core import FiberIO
from dascore.utils.io import BinaryReader
from dascore.utils.misc import optional_import

from .utils import (
    _get_attrs,
    _get_coords,
    _get_filtered_data_and_coords,
    _get_segy_version,
    _write_segy,
)


class SegyV1_0(FiberIO):  # noqa
    """An IO class supporting version 1.0 of the SEGY format."""

    name = "segy"
    preferred_extensions = ("segy", "sgy")
    # also specify a version so when version 2 is released you can
    # just make another class in the same module named JingleV2.
    version = "1.0"
    # The name of the package to import. This is here so the class can be
    # subclassed and this changed for debugging reasons.
    _package_name = "segyio"

    def get_format(self, fp: BinaryReader, **kwargs) -> tuple[str, str] | bool:
        """Make sure input is segy."""
        return _get_segy_version(fp)

    def read(self, path, time=None, channel=None, **kwargs):
        """
        Read should take a path and return a patch or sequence of patches.

        It can also define its own optional parameters, and should always
        accept kwargs. If the format supports partial reads, these should
        be implemented as well.
        """
        segyio = optional_import(self._package_name)
        with segyio.open(path, ignore_geometry=True) as fi:
            coords = _get_coords(fi)
            attrs = _get_attrs(fi, coords, path, self)
            data, coords = _get_filtered_data_and_coords(
                fi, coords, time=time, channel=channel
            )

        patch = dc.Patch(coords=coords, data=data, attrs=attrs)
        patch_trimmed = patch.select(time=time, channel=channel)
        return dc.spool([patch_trimmed])

    def scan(self, path, **kwargs) -> list[dc.PatchAttrs]:
        """
        Used to get metadata about a file without reading the whole file.

        This should return a list of
        [`PatchAttrs`](`dascore.core.attrs.PatchAttrs`) objects
        from the [dascore.core.attrs](`dascore.core.attrs`) module, or a
        format-specific subclass.
        """
        segyio = optional_import(self._package_name)
        with segyio.open(path, ignore_geometry=True) as fi:
            coords = _get_coords(fi)
            attrs = _get_attrs(fi, coords, path, self)
        return [attrs]

    def write(self, spool: dc.Patch | dc.BaseSpool, resource, **kwargs):
        """
        Create a segy file from length 1 spool or patch.

        Parameters
        ----------
        spool
            The patch or length 1 spool to write.
        resource
            The target for writing patch.

        Notes
        -----
        Based on the example from segyio:
        https://github.com/equinor/segyio/blob/master/python/examples/make-file.py
        """
        segyio = optional_import(self._package_name)
        _write_segy(spool, resource, self.version, segyio)


class SegyV0_0(SegyV1_0):  # noqa
    """
    An IO class supporting version 0.0 of the SEGY format.

    Or if the version is not set.
    """

    version = "0.0"


class SegyV0_100(SegyV1_0):  # noqa
    """
    An IO class supporting version 0.100 of the SEGY format.

    This odd version is output by some Silixa (Carina) files.
    """

    version = "0.100"


class SegyV2_0(SegyV1_0):  # noqa
    """An IO class supporting version 2.0 of the SEGY format."""

    version = "2.0"


class SegyV2_1(SegyV1_0):  # noqa
    """An IO class supporting version 2.1 of the SEGY format."""

    version = "2.1"



"""Pydantic schemas used by DASCore."""

from __future__ import annotations

import warnings
from collections.abc import Mapping
from typing import Annotated, Any, Literal

import numpy as np
from pydantic import ConfigDict, Field, PlainValidator, model_validator
from typing_extensions import Self

import dascore as dc
from dascore.constants import (
    VALID_DATA_CATEGORIES,
    VALID_DATA_TYPES,
    max_lens,
)
from dascore.core.coords import BaseCoord, CoordSummary
from dascore.utils.attrs import separate_coord_info
from dascore.utils.mapping import FrozenDict
from dascore.utils.misc import (
    to_str,
)
from dascore.utils.models import (
    CommaSeparatedStr,
    DascoreBaseModel,
    UnitQuantity,
    frozen_dict_serializer,
    frozen_dict_validator,
)

str_validator = PlainValidator(to_str)
_coord_summary_suffixes = set(CoordSummary.model_fields)
_coord_required = {"min", "max"}


def _get_coords_dict(data_dict):
    """
    Add coords dict to data dict, pop out any coordinate attributes.

    For example, if time_min, time_step are in data_dict, these will be
    grouped into the coords sub dict under "time".
    """

    def _get_dims(data_dict):
        """Try to get dim tuple."""
        dims = None
        if "dims" in data_dict:
            dims = data_dict["dims"]
        elif hasattr(coord := data_dict.get("coords"), "dims"):
            dims = coord.dims
        if isinstance(dims, str):
            dims = tuple(dims.split(","))
        return dims

    dims = _get_dims(data_dict)
    coord_info, new_attrs = separate_coord_info(
        data_dict, dims, required=("min", "max")
    )
    new_attrs["coords"] = {i: dc.core.CoordSummary(**v) for i, v in coord_info.items()}
    return new_attrs


class PatchAttrs(DascoreBaseModel):
    """
    The expected attributes for a Patch.

    The default attributes are:
    ```{python}
    #| echo: false

    import dascore as dc
    from IPython.display import Markdown

    df_str = (
        dc.PatchAttrs.get_summary_df()
        .reset_index()
        .to_markdown(index=False, stralign="center")
    )
    Markdown(df_str)
    ```
    """

    model_config = ConfigDict(
        title="Patch Summary",
        extra="allow",
        frozen=True,
        arbitrary_types_allowed=True,
    )

    data_type: Annotated[Literal[VALID_DATA_TYPES], str_validator] = Field(
        description="Describes the quantity being measured.", default=""
    )
    data_category: Annotated[Literal[VALID_DATA_CATEGORIES], str_validator] = Field(
        description="Describes the type of data.",
        default="",
    )
    data_units: UnitQuantity | None = Field(
        default=None, description="The units of the data measurements"
    )
    instrument_id: str = Field(
        description="A unique id for the instrument which generated the data.",
        default="",
        max_length=max_lens["instrument_id"],
    )
    acquisition_id: str = Field(
        description="A unique identifier linking this data to an experiment.",
        default="",
        max_length=max_lens["experiment_id"],
    )
    tag: str = Field(
        default="", max_length=max_lens["tag"], description="A custom string field."
    )
    station: str = Field(
        default="", max_length=max_lens["station"], description="A station code."
    )
    network: str = Field(
        default="", max_length=max_lens["network"], description="A network code."
    )
    history: str | tuple[str, ...] = Field(
        default_factory=tuple,
        description="A list of processing performed on the patch.",
    )

    dims: CommaSeparatedStr = Field(
        default="",
        max_length=max_lens["dims"],
        description="A tuple of comma-separated dimensions names.",
    )

    coords: Annotated[
        FrozenDict[str, CoordSummary],
        frozen_dict_validator,
        frozen_dict_serializer,
    ] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def parse_coord_attributes(cls, data: Any) -> Any:
        """Parse the coordinate attributes into coord dict."""
        if isinstance(data, dict):
            data = _get_coords_dict(data)
            # add dims as coords if dims is not included.
            if "dims" not in data:
                data["dims"] = tuple(data["coords"])
        return data

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __len__(self):
        return len(self.model_dump())

    def __getattr__(self, item):
        """Enables dynamic attributes such as time_min, time_max, etc."""
        split = item.split("_")
        # this only works on names like time_max, distance_step, etc.
        if not len(split) == 2:
            return super().__getattr__(item)
        first, second = split
        if first == "d":
            first, second = second, "step"
            msg = f"{item} is depreciated, use {first}_{second} instead."
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
        if first not in self.coords:
            return super().__getattr__(item)
        coord_sum = self.coords[first]
        return getattr(coord_sum, second)

    def get(self, item, default=None):
        """dict-like get method."""
        try:
            return self[item]
        except (AttributeError, KeyError):
            return default

    def items(self):
        """Yield (attribute, values) just like dict.items()."""
        yield from self.model_dump().items()

    def coords_from_dims(self) -> Mapping[str, BaseCoord]:
        """Return coordinates from dimensions assuming evenly sampled."""
        out = {}
        for dim in self.dim_tuple:
            out[dim] = self.coords[dim].to_coord()
        return out

    @classmethod
    def from_dict(
        cls,
        attr_map: Mapping | PatchAttrs,
    ) -> Self:
        """
        Get a new instance of the PatchAttrs.

        Optionally, give preference to data contained in a
        [`CoordManager`](`dascore.core.coordmanager.CoordManager`).

        Parameters
        ----------
        attr_map
            Anything convertible to a dict that contains the attr info.
        """
        if isinstance(attr_map, cls):
            return attr_map
        out = {} if attr_map is None else attr_map
        return cls(**out)

    @property
    def dim_tuple(self):
        """Return a tuple of dimensions. The dims attr is a string."""
        dim_str = self.dims
        if not dim_str:
            return tuple()
        return tuple(self.dims.split(","))

    def rename_dimension(self, **kwargs):
        """Rename one or more dimensions if in kwargs. Return new PatchAttrs."""
        if not (dims := set(kwargs) & set(self.dim_tuple)):
            return self
        new = self.model_dump(exclude_defaults=True)
        coords = new.get("coords", {})
        new_dims = list(self.dim_tuple)
        for old_name, new_name in {x: kwargs[x] for x in dims}.items():
            new_dims[new_dims.index(old_name)] = new_name
            coords[new_name] = coords.pop(old_name, None)
        new["dims"] = tuple(new_dims)
        return self.__class__(**new)

    def update(self, **kwargs) -> Self:
        """Update an attribute in the model, return new model."""
        coord_info, attr_info = separate_coord_info(kwargs, dims=self.dim_tuple)
        out = self.model_dump(exclude_unset=True)
        out.update(attr_info)
        out_coord_dict = out["coords"]
        for name, coord_dict in coord_info.items():
            if name not in out_coord_dict:
                out_coord_dict[name] = coord_dict
            else:
                out_coord_dict[name].update(coord_dict)
        # silly check to clear coords
        if not kwargs.get("coords", True):
            out["coords"] = {}
        return self.__class__(**out)

    def drop_private(self) -> Self:
        """Drop all private attributes."""
        contents = dict(self)
        out = {i: v for i, v in contents.items() if not i.startswith("_")}
        return self.__class__(**out)

    def flat_dump(self, dim_tuple=False, exclude=None) -> dict:
        """
        Flatten the coordinates and dump to dict.

        Parameters
        ----------
        dim_tuple
            If True, return dimensional tuple instead of range. EG, the
            output will have {time: (min, max)} rather than
            {time_min: ..., time_max: ...,}. This is useful because it can
            be passed to read, scan, select, etc.
        exclude
            keys to exclude.
        """
        out = self.model_dump(exclude=exclude)
        for coord_name, coord in out.pop("coords").items():
            names = list(coord)
            if dim_tuple:
                names = sorted(set(names) - {"min", "max"})
                out[coord_name] = (coord["min"], coord["max"])
            for name in names:
                out[f"{coord_name}_{name}"] = coord[name]
            # ensure step has right type if nullish
            step_name = f"{coord_name}_step"
            step, start = out[step_name], coord["min"]
            if step is None:
                is_time = isinstance(start, np.datetime64 | np.timedelta64)
                if is_time:
                    out[step_name] = np.timedelta64("NaT")
                elif isinstance(start, float | np.floating):
                    out[step_name] = np.nan
        return out



"""Module to transform a Patch into spectrograms."""

from __future__ import annotations

from operator import mul

import numpy as np
from scipy.signal import spectrogram as scipy_spectrogram

from dascore.constants import PatchType
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_compatible_values, get_coord
from dascore.utils.misc import iterate
from dascore.utils.patch import (
    _get_data_units_from_dims,
    _get_dx_or_spacing_and_axes,
    patch_function,
)
from dascore.utils.transformatter import FourierTransformatter


def _get_new_original_coord(old_coord, array):
    """Get a new coordinate for original axis (eg time)."""
    # determine what dtype the data
    old_min = get_compatible_values(old_coord.min(), array.dtype)
    is_dt = np.issubdtype(old_coord.dtype, np.datetime64)
    val_dtype = np.timedelta64 if is_dt else old_coord.dtype
    vals = get_compatible_values(array, val_dtype)
    return get_coord(data=old_min + vals, units=old_coord.units)


def _get_transformed_coord(coord, freqs):
    """Get the transformed coordinates."""
    units = 1 / coord.units if coord.units is not None else None
    return get_coord(data=freqs, units=units)


def _get_new_attrs(patch, cm, dim):
    """Update attributes."""
    new = dict(patch.attrs)
    new["dims"] = cm.dims
    new["data_units"] = _get_data_units_from_dims(patch, dim, mul)
    return PatchAttrs(**new)


def _get_new_dims(patch, dim, new_coord_name):
    """
    Get a new dimension tuple.

    The new dimension always takes the place of the transformed dimension,
    and the transformed dimension is appended to the end.
    EG, ("time", "distance") dimensions become ("ft_time", "distance", "time").
    """
    dims = list(patch.dims)
    dims[dims.index(dim)] = new_coord_name
    return tuple([*dims, dim])


@patch_function()
def spectrogram(patch: PatchType, dim: str = "time", **kwargs) -> PatchType:
    """
    Calculate a spectrogram from the patch data.

    The output patch will have one more dimensions than the input patch.

    Parameters
    ----------
    patch
        The input patch
    dim
        The dimension along which the spectrograms are calculated.
    **kwargs
        Passed to `scipy.signal.spectrogram` to control spectrogram options.
        See its documentation for options.

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # calculate spectrogram along time axis
    >>> time_spec = patch.spectrogram("time")
    >>> # note the new dimensions
    >>> print(time_spec.dims)
    ('distance', 'ft_time', 'time')
    >>> # perform fourier transforms along distance dimensions
    >>> dist_spec = patch.spectrogram("distance")
    """
    assert len(iterate(dim)) == 1, "only one dimension allowed."
    coord = patch.get_coord(dim)
    dxs, axes = _get_dx_or_spacing_and_axes(patch, dim, require_evenly_spaced=True)
    new_coord_name = FourierTransformatter().rename_dims(dim)[0]
    out_coords = patch.coords._get_dim_array_dict()
    # returns frequency, new values for original dimension (eg time) and spectrogram
    freqs, original, spec = scipy_spectrogram(
        patch.data,
        fs=1 / dxs[0],  # sample *frequency* is requested, not spacing
        axis=axes[0],
        **kwargs,
    )
    # add new coordinates
    out_coords[dim] = _get_new_original_coord(coord, original)
    out_coords[new_coord_name] = _get_transformed_coord(coord, freqs)
    new_dims = _get_new_dims(patch, dim, new_coord_name)
    cm = get_coord_manager(out_coords, dims=tuple(new_dims))
    attrs = _get_new_attrs(patch, cm, dim)
    return patch.__class__(data=spec, attrs=attrs, coords=cm)



"""Misc. tests for xml binary."""

from __future__ import annotations

import shutil
from pathlib import Path

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import UnknownFiberFormatError
from dascore.io.xml_binary import XMLBinaryV1
from dascore.io.xml_binary.utils import _read_xml_metadata
from dascore.utils.time import to_float

sampling_rate = 1000
expected_duration = 1

metadata = f"""<?xml version='1.0' encoding='utf-8'?>
<Metadata>
  <FileFormat>RAW</FileFormat>
  <DateTime>2024-05-30T01:15:00Z</DateTime>
  <DasInterrogatorSerial>
    <Interrogator_1>CRI-4400_A001</Interrogator_1>
  </DasInterrogatorSerial>
  <GaugeLengthM>1</GaugeLengthM>
  <PulseWidthNs>100</PulseWidthNs>
  <DataType>uint16</DataType>
  <NumberOfLasers>2</NumberOfLasers>
  <ITUChannels>
    <Laser_1>15</Laser_1>
    <Laser_2>10</Laser_2>
  </ITUChannels>
  <OriginalTemporalSamplingRate>1000</OriginalTemporalSamplingRate>
  <OutputTemporalSamplingRate>{sampling_rate}</OutputTemporalSamplingRate>
  <OriginalSpatialSamplingInterval>1</OriginalSpatialSamplingInterval>
  <Units>m</Units>
  <Zones>
    <Zone>
      <StartChannel>1</StartChannel>
      <EndChannel>10</EndChannel>
      <Stride>1</Stride>
      <NumberOfChannels>10</NumberOfChannels>
    </Zone>
  </Zones>
  <NumberOfChannels>10</NumberOfChannels>
  <NumberOfFrames>1000</NumberOfFrames>
  <UseRelativeStrain>False</UseRelativeStrain>
  <TransposedData>False</TransposedData>
</Metadata>
"""


@pytest.fixture(scope="session")
def binary_xml_directory(tmp_path_factory):
    """Creates a directory of binary files and an xml metadata."""
    new = tmp_path_factory.mktemp("xml_binary_test_data")
    metadata_path = new / "metadata.xml"
    data_1_path = new / "DAS_20240530T011500_000000Z.raw"
    data_2_path = new / "DAS_20240530T011501_000000Z.raw"
    with open(metadata_path, "w") as fi:
        fi.write(metadata)
    with open(data_1_path, "wb") as fi:
        ar = np.arange(1000 * 10, dtype=np.dtype("uint16"))
        ar.tofile(fi)
    with open(data_2_path, "wb") as fi:
        ar = np.arange(1000 * 10, dtype=np.dtype("uint16"))
        ar.tofile(fi)
    return new


@pytest.fixture(scope="session")
def binary_xml_with_other_files(
    binary_xml_directory,
    tmp_path_factory,
):
    """Ensure other files can be included at the top level."""
    new = tmp_path_factory.mktemp("binary_xml_and_others")
    shutil.copytree(binary_xml_directory, new / "xml_binary")
    sp = dc.get_example_spool()
    dc.examples.spool_to_directory(sp, new)
    return new


@pytest.fixture(scope="session")
def xml_directory_no_data(tmp_path_factory):
    """Creates a directory of xml metadata with no binary files."""
    new = tmp_path_factory.mktemp("binary_xml_no_data")
    metadata_path = new / "metadata.xml"
    with open(metadata_path, "w") as fi:
        fi.write(metadata)
    return new


@pytest.fixture(scope="session")
def directory_bad_xml(tmp_path_factory):
    """Create a directory with bad xml metadata."""
    new = tmp_path_factory.mktemp("xml_binary_bad_xml")
    metadata_path = new / "metadata.xml"
    with open(metadata_path, "w") as fi:
        fi.write("The bathhouse makes him crazy.")
    return new


class TestReadXMLMetadata:
    """Misc tests reading xml metadata."""

    def test_read_metadata_contents(self, binary_xml_directory):
        """Test reading xml metadata contents and metadata types."""
        metadata_path_1 = binary_xml_directory / "metadata.xml"
        metadata = _read_xml_metadata(metadata_path_1)
        assert metadata is not None
        # Just test a couple of attrs, pydantic should handle the rest.
        expected_attrs = ["file_format", "date_time", "units"]
        for attr in expected_attrs:
            assert hasattr(metadata, attr)


class TestGetFormat:
    """Test suite for xml binary get format."""

    def test_returns_name_format(self, binary_xml_directory):
        """Ensure the proper name/format are returned from working folder."""
        name, version = dc.get_format(binary_xml_directory)
        fiber_io = XMLBinaryV1()
        assert name == fiber_io.name
        assert version == fiber_io.version

    def test_bad_xml_metadata(self, directory_bad_xml):
        """Ensure get format raises with bad xml metadata."""
        # dc.get_format should raise, as per its docs
        with pytest.raises(UnknownFiberFormatError):
            dc.get_format(directory_bad_xml)
        # but the specific class get_format should return false.
        fiber_io = XMLBinaryV1()
        out = fiber_io.get_format(directory_bad_xml)
        assert not out


class TestScanContents:
    """Test scanning contents of xml binary directory."""

    def test_two_patches(self, binary_xml_directory):
        """Ensure the default test case has two patches."""
        fiber = XMLBinaryV1()
        out = fiber.scan(binary_xml_directory)
        assert len(out) == 2

    def test_mtime(self, binary_xml_directory):
        """Ensure scan returns contents appropriate to mtime."""
        # With no time specified, all contents should be scanned.
        scan1 = dc.scan_to_df(binary_xml_directory)
        assert len(scan1) == 2
        # When time is specified only those after should be returned.
        mtime = Path(binary_xml_directory).stat().st_mtime
        scan2 = dc.scan(binary_xml_directory, timestamp=mtime + 50)
        assert not len(scan2)
        scan3 = dc.scan(binary_xml_directory, timestamp=mtime - 50)
        assert len(scan3) == 2


class TestRead:
    """Tests for reading contents into Patches."""

    def test_read_single_file(self, binary_xml_directory):
        """Ensure we can read a single binary file in the directory."""
        fiber_io = XMLBinaryV1()
        path = next(binary_xml_directory.glob("*.raw"))
        out = fiber_io.read(path)
        assert isinstance(out, dc.BaseSpool)
        assert len(out) == 1

    def test_read_whole_directory(self, binary_xml_directory):
        """Ensure the simple path can be read by fiber io instance."""
        fiber_io = XMLBinaryV1()
        spool = fiber_io.read(binary_xml_directory)
        assert len(spool) == 2
        for patch in spool:
            assert isinstance(patch, dc.Patch)
            # test time axis
            time_coord = patch.get_coord("time")
            time_step = to_float(time_coord.step)
            assert np.allclose(time_step, 1 / sampling_rate)
            # Expecting duration to be 1 second for each file.
            duration = (time_coord.max() - time_coord.min()) / dc.to_timedelta64(1)
            assert np.isclose(duration + time_step, 1)

    def test_doesnt_reindex(self, binary_xml_directory):
        """Indexing twice shouldn't double length of spool."""
        spool = dc.spool(binary_xml_directory)
        assert len(spool) == 2
        # Before, the spool would re-index exactly once doubling size.
        # We need to ensure this doesn't happen.
        new_spool = spool.update()
        assert len(new_spool) == 2

    def test_simple_spool(self, binary_xml_directory):
        """Ensure the simple path can be read into a spool."""
        spool = dc.spool(binary_xml_directory).update()
        assert isinstance(spool, dc.BaseSpool)
        assert len(spool) == 2

    def test_read_with_other_files(self, binary_xml_with_other_files):
        """Ensure other files are also included/indexed."""
        spool = dc.spool(binary_xml_with_other_files).update()
        assert len(spool) == 5
        for patch in spool:
            assert isinstance(patch, dc.Patch)

    def test_read_empty_data_dir(self, xml_directory_no_data):
        """Try reading the directory without data."""
        fiberio = XMLBinaryV1()
        path = Path(xml_directory_no_data)
        out = fiberio.read(path)
        assert not len(out)



"""Utilities functions for Neubrex DAS IO support"""

import dascore as dc
from dascore.utils.misc import maybe_get_items, unbyte


def _is_neubrex(h5fi):
    """Determine if the file is of Neubrex origin."""
    expected_attrs = {
        "NEUBRESCOPE.DAS.Model",
        "NEUBRESCOPE.DAS.Model",
        "StartPosition",
        "StopPosition",
        "SpatialSamplingInterval",
        "GPSTimeStamp(UTC)",
        "CPUTimeStamp(UTC)",
    }
    acoustic = h5fi.get("Acoustic", None)
    attrs = set(getattr(acoustic, "attrs", set()))
    return expected_attrs.issubset(attrs)


def _get_coord_manager(acoustic):
    """Get a coordinate manager from the file."""

    def _get_time_coord(acoustic):
        """Get the time coordinate."""
        attrs = acoustic.attrs

        # We havent encountered a time decimated file yet; raise over guess
        assert attrs["TimeDecimationFilter"] in {0, 1}, "not implemented"

        gps = unbyte(attrs["GPSTimeStamp(UTC)"])
        cpu = unbyte(attrs["CPUTimeStamp(UTC)"])
        start = dc.to_datetime64(gps if gps else cpu)
        step = dc.to_timedelta64(attrs["TimeSamplingInterval(seconds)"])
        time_len = acoustic.shape[0]
        stop = start + step * time_len
        return dc.get_coord(start=start, step=step, stop=stop)

    def _get_dist_coord(acoustic):
        """Get the distance (depth) coordinate."""
        dist_len = acoustic.shape[1]
        attrs = acoustic.attrs

        # We havent encountered a distance decimated file yet; raise over guess
        assert attrs["DistanceDecimationFilter"] in {0, 1}, "not yet implemented"

        step = attrs["SpatialSamplingInterval"]
        units = dc.get_quantity(attrs["StartStopPositionUnit"])
        start = attrs["StartPosition"]
        stop = dist_len * step + start
        return dc.get_coord(start=start, step=step, units=units, stop=stop)

    coords = {
        "time": _get_time_coord(acoustic),
        "distance": _get_dist_coord(acoustic),
    }
    assert len(acoustic.shape) == 2, "Expecting 2D Neubrex array."
    return dc.get_coord_manager(coords=coords, dims=("time", "distance"))


def _get_data_units_and_type(attrs):
    """Get the units from contained string."""
    data_unit_str = unbyte(attrs["DataPhysicalUnit"]).replace("-", "")
    return dc.get_quantity(data_unit_str)


def _get_attr_dict(acoustic):
    """Get a dict of neubrex attributes."""
    mapping = {
        "GaugeLength": "gauge_length",
        "GaugeLengthUnits": "gauge_length_units",
        "IndexOfRefraction": "index_of_refraction",
        "PhaseToStrainConversion(MicroStrainPerRadian)": "phase_to_strain",
        "NEUBRESCOPE.DAS.SerialNum": "instrument_id",
        "NEUBRESCOPE.DAS.Model": "instrument_model",
        "DistanceDecimationFilter": "distance_decimation_filter",
        "TimeDecimationFilter": "time_decimation_filter",
    }
    attrs = dict(acoustic.attrs)
    out = maybe_get_items(attrs, mapping)
    out["data_units"] = _get_data_units_and_type(attrs)
    return out


def _maybe_trim_data(cm, data, time=None, distance=None, **kwargs):
    """Maybe trim the data."""
    if time is not None or distance is not None:
        cm, data = cm.select(time=time, distance=distance, array=data)
    return cm, data


def _get_attrs_coords_and_data(h5fi):
    """Return the attributes, coordinates, and data array."""
    acoustic = h5fi["Acoustic"]
    cm = _get_coord_manager(acoustic)
    attrs = _get_attr_dict(acoustic)
    return attrs, cm, acoustic



"""Tests for signal whiten."""

import numpy as np
import pytest

import dascore as dc
from dascore import get_example_patch
from dascore.exceptions import ParameterError
from dascore.units import Hz


class TestWhiten:
    """Tests for the whiten module."""

    @pytest.fixture(scope="class")
    def test_patch(self):
        """Return a shot-record used for testing patch."""
        test_patch = get_example_patch("dispersion_event")
        return test_patch.resample(time=(200 * Hz))

    def test_whiten(self, test_patch):
        """Check consistency of test_dispersion module."""
        # assert velocity dimension
        whitened_patch = test_patch.whiten(5, time=(10, 50))
        assert "distance" in whitened_patch.dims
        # assert time dimension
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )

        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_default_whiten_no_input(self, test_patch):
        """Ensure whiten can run without any input."""
        whitened_patch = test_patch.whiten()
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_default_whiten_no_smoothing_window(self, test_patch):
        """Ensure whiten can run without smoothing window size."""
        whitened_patch = test_patch.whiten(time=(5, 60))
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_smooth_window_params(self, test_patch):
        """Ensure incorrect values for smooth window raise ParameterError."""
        msg = "Frequency smoothing size must be positive"
        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(smooth_size=-1, time=[30, 60])

        msg = "Frequency smoothing size is larger than Nyquist"
        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(smooth_size=110, time=[30, 60])

    def test_default_whiten_no_freq_range(self, test_patch):
        """Ensure whiten can run without frequency range."""
        whitened_patch = test_patch.whiten(smooth_size=10)
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_edge_whiten(self, test_patch):
        """Ensure whiten can run with edge cases frequency range."""
        whitened_patch = test_patch.whiten(smooth_size=10, time=[0, 50])
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

        whitened_patch = test_patch.whiten(smooth_size=10, time=[50, 100])
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_short_windows_raises(self, test_patch):
        """Ensure too narrow frequency choices raise ParameterError."""
        msg = "Frequency range is too narrow"
        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(smooth_size=3, time=[10.02, 10.03])

    def test_bad_water_level_raises(self, test_patch):
        """Ensure bad water level values raise ParameterError."""
        msg = "water_level must be a float"

        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(water_level=[1, 2, 3], smooth_size=10)
        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(water_level=np.array([1, 2, 3]), smooth_size=10)
        with pytest.raises(ParameterError, match=msg):
            test_patch.whiten(water_level=-0.1, smooth_size=10)

    def test_whiten_monochromatic_input(self):
        """Ensures correct behavior on monochromatic signal."""
        patch = get_example_patch("sin_wav", frequency=100, sample_rate=500)
        dft_pre = patch.dft("time", real=True)

        import dascore as dc

        dc._bob = True

        white_patch = patch.whiten(smooth_size=5, time=[80, 120])
        dft_post = white_patch.dft("time", real=True)

        # Approx. symmetry for range outside frequency
        ratio_noise = np.median(
            np.abs(dft_post.select(ft_time=(120, 160)).data)
        ) / np.median(np.abs(dft_post.select(ft_time=(40, 80)).data))
        assert 0.5 < ratio_noise < 2

        # Increasing peak-to-average value in smoothing window region, right side
        post_ratio = np.median(
            np.abs(dft_post.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_post.select(ft_time=(101, 105)).data))
        pre_ratio = np.median(
            np.abs(dft_pre.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_pre.select(ft_time=(101, 105)).data))
        assert post_ratio / pre_ratio < 0.5

        # Increasing peak-to-average value in smoothing window region, left side
        post_ratio = np.median(
            np.abs(dft_post.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_post.select(ft_time=(95, 99)).data))
        pre_ratio = np.median(
            np.abs(dft_pre.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_pre.select(ft_time=(95, 99)).data))
        assert post_ratio / pre_ratio < 0.5

        # Increasing peak-to-average value in frequency range, left side
        post_ratio = np.median(
            np.abs(dft_post.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_post.select(ft_time=(105, 120)).data))
        pre_ratio = np.median(
            np.abs(dft_pre.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_pre.select(ft_time=(105, 120)).data))
        assert post_ratio / pre_ratio < 0.1

        # Increasing peak-to-average value in frequency range, right side
        post_ratio = np.median(
            np.abs(dft_post.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_post.select(ft_time=(80, 95)).data))
        pre_ratio = np.median(
            np.abs(dft_pre.select(ft_time=(99, 101)).data)
        ) / np.median(np.abs(dft_pre.select(ft_time=(80, 95)).data))
        assert post_ratio / pre_ratio < 0.1

    def test_whiten_along_distance(self, test_patch):
        """Ensure whitening runs along other axis."""
        whitened_patch = test_patch.whiten(distance=(0.001, 0.03))
        assert "distance" in whitened_patch.dims
        assert "time" in whitened_patch.dims
        assert np.array_equal(
            test_patch.coords.get_array("time"), whitened_patch.coords.get_array("time")
        )
        assert np.array_equal(
            test_patch.coords.get_array("distance"),
            whitened_patch.coords.get_array("distance"),
        )

    def test_whiten_dft_input(self, test_patch):
        """
        Ensure whiten function returns dft patch when dft patch is input.
        """
        dft = test_patch.dft("time", real=True)
        whitened_patch_freq_domain = dft.whiten(smooth_size=5, time=None)

        # check if the returned data is in the frequency domain
        assert np.iscomplexobj(
            whitened_patch_freq_domain.data
        ), "Expected the output to be complex, indicating freq. domain patch."

        assert "ft_time" in dft.coords.coord_map

    def test_whiten_df_all_parameters(self, test_patch):
        """Ensure whiten accepts all args in dft form."""
        fft_patch = test_patch.dft("time", real=True)

        whitened_patch_freq_domain = fft_patch.whiten(
            smooth_size=5, time=(30, 60), water_level=0.01
        )
        assert isinstance(whitened_patch_freq_domain, dc.Patch)

    def test_no_time_no_kwargs_raises(self, random_patch):
        """Ensure if no kwargs and patch doesn't have time an error is raised."""
        patch = random_patch.rename_coords(time="money")
        msg = "and patch has no time dimension"
        with pytest.raises(ParameterError, match=msg):
            patch.whiten()

    def test_bad_dim_name_in_kwargs_raises(self, random_patch):
        """Ensure a bad dimension name raises."""
        msg = "whiten but it is not in patch dimensions"
        with pytest.raises(ParameterError, match=msg):
            random_patch.whiten(bad_dim=(1, 10))

    def test_multiple_kwargs_raises(self, random_patch):
        """Ensure passing multiple kwargs raises error."""
        msg = "must specify a single patch dimension"
        with pytest.raises(ParameterError, match=msg):
            random_patch.whiten(time=None, distance=None)

    def test_helpful_error_message(self, random_patch):
        """Ensure helpful error message is used when a bad kwarg is passed."""



"""
SEGY format support module.

Notes
-----
- Distance information is not found in most SEGY DAS files so returned
  dimensions are "channel" and "time" rather than "distance" and "time".
- Segy standards found at: https://library.seg.org/pb-assets/technical-standards

segy v1 spec: seg_y_rev1-1686080991247.pdf

segy v2 spec: seg_y_rev2_0-mar2017-1686080998003.pdf

segy v2.1 spec: seg_y_rev2_1-oct2023-1701361639333.pdf

Examples
--------
import dascore as dc
from dascore.utils.downloader import fetch

# get the path to a segy file.
path = fetch("conoco_segy_1.sgy")

segy_patch = dc.spool(path)[0]
"""

from .core import SegyV1_0, SegyV2_0, SegyV2_1



"""Module for waterfall plotting."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Literal

import matplotlib.pyplot as plt
import numpy as np

from dascore.constants import PatchType
from dascore.utils.patch import patch_function
from dascore.utils.plotting import (
    _get_ax,
    _get_cmap,
    _get_dim_label,
)


def _set_scale(im, scale, scale_type, color_coords):
    """Set the scale of the color bar based on scale and scale_type."""
    # check scale paramters
    assert scale_type in {"absolute", "relative"}
    assert isinstance(scale, float | int) or len(scale) == 2
    # make sure we have a len two array
    modifier = 1
    if scale_type == "relative":
        modifier = 0.5 * (np.nanmax(color_coords) - np.nanmin(color_coords))
        # only one scale parameter provided, center around mean
    if isinstance(scale, float):
        mean = np.nanmean(color_coords)
        scale = np.array([mean - scale * modifier, mean + scale * modifier])
    im.set_clim(scale)


@patch_function()
def map_fiber(
    patch: PatchType,
    x: np.ndarray | str = "distance",
    y: np.ndarray | str = "distance",
    color: np.ndarray | str = "distance",
    ax: plt.Axes | None = None,
    cmap="cividis_r",
    scale: float | Sequence[float] | None = None,
    scale_type: Literal["relative", "absolute"] = "relative",
    show=False,
) -> plt.Axes:
    """
    Create a plot of the outline of the cable colorized by a given parameter.

    Parameters
    ----------
    patch
        The Patch object.
    x
        x coordinate: can be an array or a str representing a patch coordinate.
    y
        y coordinate: can be an array or a str representing a patch coordinate.
    color
        The color parameter to plot: can be an array or a str representing a patch
        attribute.
    ax
        A matplotlib object, if None create one.
    cmap
        A matplotlib colormap string or instance. Set to None to not plot the
        colorbar.
    scale
        If not None, controls the saturation level of the colorbar.
        Values can either be a float, to set upper and lower limit to the same
        value centered around the mean of the data, or a length 2 tuple
        specifying upper and lower limits. See `scale_type` for controlling how
        values are scaled.
    scale_type
        Controls the type of scaling specified by `scale` parameter. Options
        are:
            relative - scale based on half the dynamic range in patch
            absolute - scale based on absolute values provided to `scale`
    show
        If True, show the plot, else just return axis.

    Examples
    --------
    >>> # Plot patch
    >>> import dascore as dc
    >>> patch = dc.get_example_patch("random_patch_with_lat_lon")
    >>> patch = patch.set_units(latitude="m", longitude="m")
    >>> _ = patch.viz.map_fiber("latitude", "longitude", "distance")
    """
    dims = []
    if isinstance(x, str):
        assert x in patch.coords, f"{x} not found in patch coordinates"
        dims.append(x)
        x = patch.coords.get_array(x)
    if isinstance(y, str):
        assert y in patch.coords, f"{y} not found in patch coordinates"
        dims.append(y)
        y = patch.coords.get_array(y)
    if isinstance(color, str):
        assert color in patch.coords, f"{color} not found in patch coordinates"
        data_type = color
        data_units = patch.attrs.coords[color].units
        color = patch.coords.get_array(color)
    else:
        data_type = ""
        data_units = ""

    ax = _get_ax(ax)
    cmap = _get_cmap(cmap)

    im = ax.scatter(x, y, c=color, cmap=cmap)

    # scale colorbar
    if scale is not None:
        _set_scale(im, scale, scale_type, color)

    # set axis labels
    for dim, x in zip(dims, ["x", "y"]):
        getattr(ax, f"set_{x}label")(_get_dim_label(patch, dim))

    # add color bar with title
    if cmap is not None:
        cb = ax.get_figure().colorbar(im, ax=ax, fraction=0.05, pad=0.025)
        dunits = f" ({data_units})" if (data_type and data_units) else f"{data_units}"
        label = f"{data_type}{dunits}"
        cb.set_label(label)

    if show:
        plt.show()
    return ax



"""Utilities for Sentek data format."""

from __future__ import annotations

from pathlib import Path

import numpy as np

import dascore as dc
from dascore.core import get_coord, get_coord_manager


def _get_version(fid):
    """Determine if Sentek file."""
    name = fid.name
    # Sentek files cannot change the extension, or file name.
    sw_data = name.endswith(".das")
    fid.seek(0)
    # There isn't anything in the header particularly useful for determining
    # if it is a Sentek file, so we do what we can here.
    # First check if sensor_num and measurement_count are positive and nearly
    # ints.
    sensor_num = np.fromfile(fid, dtype=np.float32, count=1)[0]
    measurement_count = np.fromfile(fid, dtype=np.float32, count=1)[0]
    _ = np.fromfile(fid, dtype=np.float32, count=1)[0]  # sampling_interval
    is_positive = (sensor_num > 1) and (measurement_count > 1)
    sens_nearly_int = np.round(sensor_num, 5) == np.round(sensor_num)
    meas_nearly_int = np.round(measurement_count, 5) == np.round(measurement_count)
    nearly_ints = sens_nearly_int and meas_nearly_int
    # Then check if strain_rate value is valid.
    strain_rate = int(np.fromfile(fid, dtype=np.float32, count=1)[0])
    proper_strain_rate = strain_rate in {0, 1}
    # Note: We will need to modify this later for different versions of the
    # sentek data, but for now we only support 5.
    if sw_data and is_positive and proper_strain_rate and nearly_ints:
        return ("sentek", "5")
    return False


def _get_time_from_file_name(name) -> np.datetime64:
    """Extract time contained in the file name.

    example file name: DASDMSShot00_20230328155652124.das
    """
    time_str = name.split("_")[1].split(".")[0]
    year = time_str[:4]
    month = time_str[4:6]
    day = time_str[6:8]
    hour = time_str[8:10]
    minute = time_str[10:12]
    second = float(time_str[12:]) / 1_000
    iso = f"{year}-{month}-{day}T{hour}:{minute}:{second:02f}"
    return np.datetime64(iso)


def _get_patch_attrs(fid, extras=None):
    """Extracts patch metadata.

    A few important fields in the header and their meaning:

    sensor_num: number of channels in the sensing fiber
    measurement_count: number of measurements in ONE single file
    sampling_interval: sampling interval in nanosecond (delta t)
    strain_rate: flag that is set when the loaded data represents strain rate
    trigger_position: index position where the trigger occurs
    decimation_factor: decimation factor (integer)
    """
    fid.seek(0)
    sensor_num = np.fromfile(fid, dtype=np.float32, count=1)[0]
    measurement_count = np.fromfile(fid, dtype=np.float32, count=1)[0]
    _ = np.fromfile(fid, dtype=np.float32, count=1)[0]  # sampling_interval
    strain_rate = np.fromfile(fid, dtype=np.float32, count=1)[0]
    _ = np.fromfile(fid, dtype=np.float32, count=1)[0]  # trigger_position
    _ = np.fromfile(fid, dtype=np.float32, count=1)[0]  # decimation_factor
    # create distance coordinate
    distance_start = np.fromfile(fid, dtype=np.float32, count=1)[0]
    fid.seek(int(sensor_num - 1) * 4)
    distance_stop = np.fromfile(fid, dtype=np.float32, count=1)[0]
    distance_step = (distance_stop - distance_start) / sensor_num
    dist = get_coord(start=distance_start, stop=distance_stop, step=distance_step)
    # create time coord
    file_time = _get_time_from_file_name(Path(fid.name).name)
    offset_start = np.fromfile(fid, dtype=np.float32, count=1)[0]
    fid.seek(int(measurement_count - 1) * 4)
    offset_stop = np.fromfile(fid, dtype=np.float32, count=1)[0]
    time_start = file_time + dc.to_timedelta64(offset_start)
    time_stop = file_time + dc.to_timedelta64(offset_stop)
    time_step = (time_stop - time_start) / measurement_count
    time = get_coord(start=time_start, stop=time_stop, step=time_step)

    data_type = "strain_rate" if strain_rate else "strain"
    coord_manager = get_coord_manager(
        {"time": time, "distance": dist}, dims=("distance", "time")
    )
    attrs = dc.PatchAttrs(
        coords=coord_manager, data_type=data_type, **({} if extras is None else extras)
    )
    offsets = fid.tell(), int(measurement_count), int(sensor_num)
    return attrs, coord_manager, offsets



"""
Module for Fourier transforms.

See the [FFT note](/notes/dft_notes.qmd) for discussion on the
implementation.
"""

from __future__ import annotations

from collections.abc import Sequence
from operator import mul

import numpy as np
import numpy.fft as nft

from dascore.constants import PatchType
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_coord
from dascore.units import invert_quantity
from dascore.utils.misc import iterate
from dascore.utils.patch import (
    _get_data_units_from_dims,
    _get_dx_or_spacing_and_axes,
    patch_function,
)
from dascore.utils.time import to_float
from dascore.utils.transformatter import FourierTransformatter


def _get_dft_new_coords(patch, dxs, dims, axes, real, original_cm=None):
    """
    Create coordinates based on dxs and patch shape.

    if original_cm is not none, it means the patch was padded.
    """
    # Note: We need original_cm and patch because patch may have undergone
    # padding.

    def _get_fft_coord(x_len, dx, units):
        """Get coord for normal fft coord."""
        new_dx = 1.0 / (x_len * dx)
        stop = ((x_len - 1) // 2 + 1) * new_dx
        start = -(x_len // 2) * new_dx
        units = invert_quantity(units)
        return get_coord(start=start, stop=stop, step=new_dx, units=units)

    def _get_rfft_coord(x_len, dx, units):
        """Get coord from real fft coord."""
        new_dx = 1.0 / (x_len * dx)
        start = 0
        stop = (x_len // 2 + 1) * new_dx
        units = invert_quantity(units)
        return get_coord(start=start, stop=stop, step=new_dx, units=units)

    # first disassociate old coordinates. We do this rather than drop them
    # so the idft can find them and exactly restore old coords.
    old_cm = patch.coords.disassociate_coord(*dims)
    new_coords = old_cm.get_coord_tuple_map()
    ft = FourierTransformatter()
    for i, dim in enumerate(dims):
        old_coord = patch.get_coord(dim)
        units = old_coord.units
        size = old_coord.shape[0]
        dx = dxs[i]
        new_name = ft.rename_dims(dim)[0]
        if dim == real:
            coord = _get_rfft_coord(size, dx, units)
        else:
            coord = _get_fft_coord(size, dx, units)
        new_coords[new_name] = (new_name, coord)
        # Add padded coordinates
        if original_cm is not None:
            new_coords[f"_{dim}_unpadded"] = (None, original_cm.get_coord(dim))
    new_dims = ft.rename_dims(patch.dims, index=axes)
    cm = get_coord_manager(new_coords, dims=new_dims)
    return cm


def _get_dft_attrs(patch, dims, new_coords, pad=False):
    """Get new attributes for transformed patch."""
    new = dict(patch.attrs)
    new["dims"] = new_coords.dims
    new["data_units"] = _get_data_units_from_dims(patch, dims, mul)
    # As per #390, we also want to remove data_type (eg the patch is no
    # longer in strain rate after the dft)
    new["_pre_dft_data_type"] = new.pop("data_type", None)
    new["_dft_padded"] = pad
    return PatchAttrs(**new)


def _get_untransformed_dims(patch, dims):
    """Return dimensions which have not been transformed."""
    dim_set = set(patch.dims)
    out = []
    for dim in dims:
        # This dim has already been transformed.
        if (dim not in dim_set) and f"ft_{dim}" in dim_set:
            continue
        out.append(dim)
    return out


@patch_function()
def dft(
    patch: PatchType,
    dim: str | None | Sequence[str],
    *,
    real: str | bool | None = None,
    pad: bool = True,
) -> PatchType:
    """
    Perform the discrete Fourier transform (dft) on specified dimension(s).

    Parameters
    ----------
    patch
        Patch to transform.
    dim
        A single, or multiple dimensions over which to perform dft. If
        None, perform dft over all dimensions.
    real
        Either 1) The name of the axis over which to perform a rfft, 2)
        True, which means the last (possibly only) dimenson should have an
        rfft performed, or 3) None, meaning no rfft.
    pad
        If True, pad patch before peforming dft along desired dimensions to
        the next fast length. This can avoid major slow-downs when dimension
        lengths are prime numbers.

    Notes
    -----
    - Simply uses numpy's fft module but outputs are scaled by the sample
      spacing along each transformed dimension and coordinates corresponding
      to frequency bins are shifted so they remain ordered.

    - Each transformed dimension is renamed with a preceding `ft_`. e.g.,
      `time` becomes `ft_time` (ft stands for fourier transform).

    - Each transformed dimension has units of 1/original units.

    - Output data units are the original data units multiplied by the units
      of each transformed dimension.

    - Non-dimensional coordiantes associated with transformed coordinates
      will be dropped in the output.

    - See the [FFT notes](`docs/notes/dft_notes.qmd`) for more details.

    See Also
    --------
    -[idft](`dascore.transform.fourier.idft`)

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # perform dft (fft) on time axis
    >>> dft_time = patch.dft(dim="time")
    >>> # make it a real fft (no negative frequencies)
    >>> dft_time_real = patch.dft(dim="time", real=True)
    >>> # dft on specified dimensions, specify real dimension
    >>> dft_some_real = patch.dft(dim=("time", "distance"), real="time")
    """
    dims = list(iterate(dim if dim is not None else patch.dims))
    patch.check_coords(coords=dims)
    real = dims[-1] if real is True else real  # if true grab last dim
    dims = _get_untransformed_dims(patch, dims)
    real = real if real in dims else None  # may need to reset real
    if not dims:  # no transformation needed.
        return patch
    # re-arrange list so real dim is last (if provided)
    if isinstance(real, str):
        assert real in dims, "real must be in provided dimensions."
        dims.append(dims.pop(dims.index(real)))
    original_cm = patch.coords if pad else None
    if pad:  # apply padding to avoid slow dft lengths.
        pad_kwargs = {x: "fft" for x in dims}
        patch = patch.pad.func(patch, **pad_kwargs)
    # get axes and spacing along desired dimensions.
    dxs, axes = _get_dx_or_spacing_and_axes(patch, dims, require_evenly_spaced=True)
    # get new coordinates (need before pad)
    new_coords = _get_dft_new_coords(
        patch, dxs, dims, axes, real, original_cm=original_cm
    )
    func = nft.rfftn if real is not None else nft.fftn
    # scale as explained above and in notes, then shift
    scale_factor = np.prod(dxs)
    fft_data = func(patch.data, axes=axes) * scale_factor
    shift_slice = slice(None) if real is None else slice(None, -1)
    data = nft.fftshift(fft_data, axes=axes[shift_slice])
    # get attributes
    attrs = _get_dft_attrs(patch, dims, new_coords, pad=pad)
    return patch.new(data=data, coords=new_coords, attrs=attrs)


def _get_idft_dims_steps_axis(patch, dim):
    """
    Get the dimensions, step sizes as a float, axis numbers and if an
    irft should be performed.
    """
    ft = FourierTransformatter()
    if dim is None:
        dim = [x for x in patch.dims if x.startswith("ft_")]
    # try to get pre-transformed names if used. EG "time" might refer to
    # ft_time for brevity.
    current_dims = set(patch.dims)
    dims = [x if x in current_dims else ft.rename_dims(x)[0] for x in iterate(dim)]
    patch.check_coords(dims=dims)
    coords = [patch.get_coord(x, require_evenly_sampled=True) for x in dims]
    is_real = [1 if to_float(x.min()) == 0 else 0 for x in coords]
    real_sum = sum(is_real)
    assert real_sum <= 1, "only one real axis allowed."
    has_real = bool(real_sum)
    # we need to move the real dim to the end of the list
    if has_real:
        real_ind = is_real.index(1)
        dims.append(dims.pop(real_ind))
    steps, axis = _get_dx_or_spacing_and_axes(patch, dims)
    return dims, steps, axis, has_real


def _get_idft_coords_and_sizes(patch, dims, new_dims, axes, real):
    """Get the new coords for the idft and expected sizes to pass to numpy."""
    shapes = patch.shape
    padded = patch.attrs.get("_dft_padded", False)
    coord_map = patch.coords.disassociate_coord(*dims).get_coord_tuple_map()
    sizes = []
    padding = {}
    for old_dim, new_dim, ax in zip(dims, new_dims, axes):
        # if old dim is stored
        ax_len = shapes[ax]
        potential_coord = coord_map.get(new_dim, (None, None))[1]
        if potential_coord is None:
            msg = (
                "Currently, IDFT can only be performed on patches which have"
                " been transformed to Fourier domain with dft method."
            )
            raise NotImplementedError(msg)
        if (len(potential_coord) == ax_len) or (real and old_dim == dims[-1]):
            sizes.append(len(potential_coord))
        coord_map[new_dim] = (new_dim, potential_coord)
        if not padded:  # No padding, go to next dim.
            continue
        old_len = len(coord_map.pop(f"_{new_dim}_unpadded")[1])
        diff = old_len - len(coord_map[new_dim][1])
        if diff < 0:
            padding[new_dim] = (0, diff)
    ft = FourierTransformatter()
    new_dims = ft.rename_dims(patch.dims, index=axes, forward=False)
    cm = get_coord_manager(coord_map, dims=new_dims).drop_coords(*dims)[0]
    out_size = np.asarray(sizes) if len(sizes) else None
    return cm, out_size, padding


def _get_idft_attrs(patch, dims, new_coords):
    """Get new attributes for transformed patch."""
    # add all {dim}_min to new coords to ensure reverse ft can restore dims.
    new = dict(patch.attrs)
    new.pop("coords", None)
    new["dims"] = new_coords.dims
    new["data_units"] = _get_data_units_from_dims(patch, dims, mul)
    # Restore the pre-dft datatype.
    if "_pre_dft_data_type" in new:
        new["data_type"] = new.pop("_pre_dft_data_type", None)
    new.pop("_dft_padded", None)
    return PatchAttrs(**new)


@patch_function()
def idft(patch: PatchType, dim: str | None | Sequence[str] = None) -> PatchType:
    """
    Perform the inverse discrete Fourier transform (idft) on specified dimension(s).

    Currently, only patches that have been transformed with
    [dft](`dascore.transform.fourier.dft`) can be used with this function.
    After transformation with dft, the transformed coordinates cannot change
    (e.g., with [select]('dascore.proc.basic.select`) otherwise idft won't
    work.

    Parameters
    ----------
    patch
        Patch to transform.
    dim
        A single, or multiple dimensions over which to perform idft. If
        None, perform idft over all dimensions that have names starting
        with "ft_", which indicates they have already undergone a fourier
        transform.

    Notes
    -----
    - Real transforms are determined by transformed coordinates which have
      no negative values.

    - See the [FFT note](dascore.org/notes/fft_notes.html) in Notes section
      of DASCore's documentation.

    See Also
    --------
    -[dft](`dascore.transform.fourier.dft`)

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # perform dft (fft) on time axis
    >>> dft_time = patch.dft(dim="time")
    >>> # get inverse dft, transformed axis are ascertained automatically
    >>> idft = dft_time.idft()
    """
    dims, steps, axes, real = _get_idft_dims_steps_axis(patch, dim)
    new_dims = FourierTransformatter().rename_dims(dims, forward=False)
    func = nft.irfftn if real else nft.ifftn
    # Get new coords, fft sizes, and padding to remove.
    coords, sizes, padding = _get_idft_coords_and_sizes(
        patch, dims, new_dims, axes, real
    )
    # now unshift data and undo scaling
    ax_slice = slice(None, -1) if real else slice(None)
    scale_factor = np.prod([to_float(coords.coord_map[x].step) for x in new_dims])
    _preped = nft.ifftshift(patch.data / scale_factor, axes=axes[ax_slice])
    data = func(_preped, s=sizes, axes=axes)
    attrs = _get_idft_attrs(patch, dims, coords)
    out = patch.new(data=data, attrs=attrs, coords=coords)
    if padding:
        out = out.select(**padding, samples=True)
    return out



"""Tests for filters."""

from __future__ import annotations

import re
import sys

import numpy as np
import pandas as pd
import pytest

import dascore as dc
from dascore.exceptions import (
    CoordDataError,
    FilterValueError,
    ParameterError,
    PatchCoordinateError,
    UnitError,
)
from dascore.units import Hz, convert_units, get_unit, m
from dascore.utils.misc import broadcast_for_index


class TestPassFilterChecks:
    """Test that bad filter checks raise appropriate errors."""

    def test_no_kwargs_raises(self, random_patch):
        """Ensure ValueError is raised with no dim parameters."""
        with pytest.raises(FilterValueError):
            _ = random_patch.pass_filter()

    def test_wrong_kwarg_length_raises(self, random_patch):
        """Only len two sequence is allowed."""
        with pytest.raises(FilterValueError, match="length two sequence"):
            _ = random_patch.pass_filter(time=[1])
        with pytest.raises(FilterValueError, match="length two sequence"):
            _ = random_patch.pass_filter(time=[1, 3, 3])

    def test_all_null_kwarg_raises(self, random_patch):
        """There must be one Non-null kwarg."""
        with pytest.raises(FilterValueError, match="at least one filter"):
            _ = random_patch.pass_filter(time=[None, np.nan])

    def test_unordered_params(self, random_patch):
        """Ensure a low parameter greater than a high parameter raises."""
        with pytest.raises(FilterValueError):
            _ = random_patch.pass_filter(time=[10, 1])

    def test_bad_low_param(self, random_patch):
        """Ensure a low parameter above nyquist raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(time=[1000, None])

    def test_bad_high_param(self, random_patch):
        """Ensure a high parameter above nyquist raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(distance=[None, 10])

    def test_filt_param_less_than_0(self, random_patch):
        """Ensure a high parameter above nyquist raises."""
        with pytest.raises(FilterValueError, match="possible filter bounds"):
            _ = random_patch.pass_filter(distance=[None, -10])

    def test_bad_units_error(self, random_patch):
        """Ensure incompatible units raise an Error."""
        m = dc.units.get_quantity("m")
        with pytest.raises(UnitError):
            random_patch.pass_filter(time=(1 * m, 10 * m))

    def test_units_dont_match(self, random_patch):
        """Tests for Units matching."""
        sec = dc.get_quantity("s")
        m = dc.get_quantity("m")
        filt = (1 * m, 10 * sec)
        match = "Units must match"
        with pytest.raises(UnitError, match=match):
            random_patch.pass_filter(time=filt)

    def test_high_time_raises(self, random_patch):
        """Ensure too high freq band in time axis raises."""
        nyquest = 0.5 / (random_patch.attrs.time_step / dc.to_timedelta64(1))
        hz = dc.get_quantity("Hz")
        filt = (1 * hz, nyquest * 1.1 * hz)
        match = "possible filter bounds are"
        with pytest.raises(FilterValueError, match=match):
            random_patch.pass_filter(time=filt)


class TestPassFilter:
    """Simple tests to make sure filter logic runs."""

    def test_time_bandpass_runs(self, random_patch):
        """Ensure filtering along time_axis works."""
        out = random_patch.pass_filter(time=(10, 100))
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_time_lowpass_runs(self, random_patch):
        """Ensure stopfilter on time works."""
        out = random_patch.pass_filter(time=(None, 100.22))
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_time_highpass_runs(self, random_patch):
        """Ensure a highpass on time runs."""
        out = random_patch.pass_filter(time=(10.2, None))
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_apply_distance_filter(self, random_patch):
        """Apply bandpass along distance dimension."""
        out = random_patch.pass_filter(distance=(0.1, 0.2))
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_uneven_sampling_raises(self, wacky_dim_patch):
        """A nice error message should be raised if the samples arent even."""
        match = "is not evenly sampled"
        with pytest.raises(CoordDataError, match=match):
            wacky_dim_patch.pass_filter(time=(10, 100))

    def test_specify_units_simple(self, random_patch):
        """Ensure units can be specified in patch arguments."""
        hz = dc.units.get_quantity("Hz")
        out1 = random_patch.pass_filter(time=(1 * hz, 10 * hz))
        out2 = random_patch.pass_filter(time=(1, 10))
        assert out1.equals(out2)

    def test_specify_units_inverse(self, random_patch):
        """Passing 1/Hz (eg seconds) should also work."""
        s = dc.units.get_quantity("s")
        out1 = random_patch.pass_filter(time=(1 * s, 10 * s))
        out2 = random_patch.pass_filter(time=(0.1, 1))
        assert out1.equals(out2)

    def test_misc_units(self, random_patch):
        """Catchall for other unit tests."""
        patch = random_patch
        m, ft = dc.units.get_unit("m"), dc.units.get_unit("ft")
        hz = dc.units.get_unit("Hz")
        # Filter from 1 Hz to 10 Hz in time dimension
        patch.pass_filter(time=(1 * hz, 10 * hz))
        # Filter wavelengths 50m to 100m
        patch.pass_filter(distance=(50 * m, 100 * m))
        # filter wavelengths less than 200 ft
        patch.pass_filter(distance=(200 * ft, None))

    def test_one_unit_raises(self, random_patch):
        """When only one unit is specified it should raise."""
        match = "Both inputs must be "
        s = dc.units.get_quantity("1")
        with pytest.raises(UnitError, match=match):
            random_patch.pass_filter(time=(1 * s, 10 * s))

    def test_ellipses(self, random_patch):
        """Ellipses should work for filtering."""
        p = random_patch
        p.pass_filter(time=(..., 20))
        assert p.pass_filter(time=(None, 20)) == p.pass_filter(time=(..., 20))
        assert p.pass_filter(time=(10, None)) == p.pass_filter(time=(10, ...))

    def test_non_zero_phase(self, random_patch):
        """Ensure non-zero-phase filter logic runs."""
        out = random_patch.pass_filter(time=(..., 20), zerophase=False)
        assert isinstance(out, dc.Patch)


class TestSobelFilter:
    """Simple tests to make sure Sobel filter runs."""

    def test_invalid_mode(self, random_patch):
        """Ensure ValueError is raised with an invalid mode."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="time", mode="test", cval=0.0)

    def test_invalid_mode_type(self, random_patch):
        """Ensure ValueError is raised with an invalid mode type."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="time", mode=0.0)

    def test_invalid_dim(self, random_patch):
        """Ensure ValueError is raised with an invalid dim type."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim=-1)

    def test_invalid_axis(self, random_patch):
        """Ensure ValueError is raised with an invalid axis value."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim=None, mode="constant", cval=0.0)

    def test_invalid_cval(self, random_patch):
        """Ensure ValueError is raised with an invalid cval value."""
        with pytest.raises(FilterValueError):
            _ = random_patch.sobel_filter(dim="distance", mode="constant", cval=None)

    def test_sobel_runs(self, random_patch):
        """Ensure Sobel filter works with default params."""
        out = random_patch.sobel_filter(dim="time")
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))


class TestMedianFilter:
    """Simple tests on median filter."""

    def test_median_no_kwargs_raises(self, random_patch):
        """Apply default values."""
        msg = "You must use exactly one"
        with pytest.raises(PatchCoordinateError, match=msg):
            random_patch.median_filter()

    def test_median_filter_time(self, random_patch):
        """Test median filter in time dimension."""
        out = random_patch.median_filter(time=0.5)
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_median_filter_time_distance(self, random_patch):
        """Apply default values."""
        out = random_patch.median_filter(time=0.05, distance=2)
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_median_filter_ones(self, random_patch):
        """Apply default values."""
        out = random_patch.median_filter(time=1, distance=1, samples=True)
        assert out == random_patch


class TestNotchFilter:
    """Tests for the notch filter."""

    def test_notch_no_kwargs_raises(self, random_patch):
        """Test that no dimension raises an appropriate error."""
        msg = "You must use exactly one"
        with pytest.raises(PatchCoordinateError, match=msg):
            random_patch.notch_filter(q=30)

    def test_notch_filter_time(self, random_patch):
        """Test the notch filter along the time dimension."""
        filtered_patch = random_patch.notch_filter(time=60, q=30)
        assert isinstance(filtered_patch, dc.Patch)
        assert not np.any(np.isnan(filtered_patch.data))

    def test_notch_filter_distance(self, random_patch):
        """Test the notch filter along the distance dimension."""
        filtered_patch = random_patch.notch_filter(distance=0.2, q=20)
        assert isinstance(filtered_patch, dc.Patch)
        assert not np.any(np.isnan(filtered_patch.data))

    def test_notch_filter_time_distance(self, random_patch):
        """Test the notch filter along the time and distance dimension."""
        filtered_patch = random_patch.notch_filter(distance=0.25, time=12, q=40)
        assert isinstance(filtered_patch, dc.Patch)
        assert not np.any(np.isnan(filtered_patch.data))

    def test_notch_filter_high_frequency_error(self, random_patch):
        """Test notch filter raises error for frequency beyond Nyquist."""
        sr = dc.utils.patch.get_dim_sampling_rate(random_patch, "time")
        nyquist = 0.5 * sr
        too_high_freq = nyquist + 1
        msg = f"possible filter values are in [0, {nyquist}] you passed {too_high_freq}"
        with pytest.raises(FilterValueError, match=re.escape(msg)):
            random_patch.notch_filter(time=too_high_freq, q=30)

    def test_notch_filter_time_units(self, random_patch):
        """Test notch filter with time dimension and frequency in Hz."""
        filtered_patch = random_patch.notch_filter(time=60 * Hz, q=40)
        assert isinstance(filtered_patch, dc.Patch)
        assert not np.any(np.isnan(filtered_patch.data))

    def test_notch_filter_distance_units(self, random_patch):
        """Test notch filter with distance dimension in meters."""
        filtered_patch = random_patch.notch_filter(distance=0.4 * 1 / m, q=25)
        assert isinstance(filtered_patch, dc.Patch)
        assert not np.any(np.isnan(filtered_patch.data))


class TestSavgolFilter:
    """Simple tests on Savgol filter."""

    def test_savgol_no_kwargs_raises(self, random_patch):
        """Apply default values."""
        msg = "You must use exactly one"
        with pytest.raises(PatchCoordinateError, match=msg):
            random_patch.savgol_filter(polyorder=2)

    def test_savgol_filter_time(self, random_patch):
        """Test savgol filter in time dimension."""
        out = random_patch.savgol_filter(polyorder=2, time=5)
        assert isinstance(out, dc.Patch)
        assert not np.any(pd.isnull(out.data))

    def test_savgol_filter_ones(self, random_patch):
        """Apply default values."""
        out = random_patch.savgol_filter(polyorder=2, distance=5, samples=True)
        assert out != random_patch

    def test_savgol_smoothing(self, random_patch):
        """Test smoothing in one dimension."""
        new_array = np.array(random_patch.data)
        midpoint = new_array.shape[0] // 2
        new_array[:midpoint] = 1
        new_array[midpoint:] = 0
        new_patch = random_patch.new(data=new_array)
        dim = new_patch.dims[0]
        out = new_patch.savgol_filter(polyorder=2, samples=True, **{dim: 5})
        assert np.allclose(out.data[:5], 1)
        assert np.allclose(out.data[-5:], 0)
        middle = out.data[midpoint - 10 : midpoint + 10]
        assert np.any((middle < 1) & (middle > 0))

    def test_savgol_filter_multiple_dims(self, event_patch_2):
        """Ensure multiple dimensions can be filtered."""
        out = event_patch_2.savgol_filter(distance=10, time=0.001, polyorder=4)
        assert out.shape == event_patch_2.shape
        assert not np.allclose(out.data, event_patch_2.data)


class TestGaussianFilter:
    """Test the Guassian Filter."""

    def test_filter_time(self, event_patch_2):
        """Test for simple filter along the time axis."""
        out = event_patch_2.gaussian_filter(time=0.001)
        assert isinstance(out, dc.Patch)
        assert out.shape == event_patch_2.shape

    def test_filter_distance(self, event_patch_2):
        """Ensure filter can be applied along distance axis with samples."""
        out = event_patch_2.gaussian_filter(distance=5, samples=True)
        assert isinstance(out, dc.Patch)
        assert out.shape == event_patch_2.shape

    def test_filter_time_distance(self, event_patch_2):
        """Ensure both time and distance can be filtered."""
        out = event_patch_2.gaussian_filter(time=5, distance=5, samples=True)
        assert isinstance(out, dc.Patch)
        assert out.shape == event_patch_2.shape


class TestSlopeFilter:
    """Test suite for slope filter."""

    def get_slope_array(self, patch, dims=("ft_time", "ft_distance")):
        """Get an array of slopes values for the patch ."""
        dim1, dim2 = dims
        dims = patch.dims
        ndims = patch.ndim
        coord1 = patch.get_array(dim1)
        coord2 = patch.get_array(dim2) + sys.float_info.epsilon
        # Need to add appropriate blank dims to keep overall shape of patch.
        ax1, ax2 = dims.index(dim1), dims.index(dim2)
        shape_1 = broadcast_for_index(ndims, ax1, value=slice(None), fill=None)
        shape_2 = broadcast_for_index(ndims, ax2, value=slice(None), fill=None)
        # Then just allow broadcasting to do its magic
        return coord1[shape_1] / coord2[shape_2]

    @pytest.fixture
    def example_patch(self, event_patch_1):
        """Return the example patch ready for slope filter."""
        out = (
            event_patch_1.taper(time=0.05)
            .pass_filter(time=(1, 500))
            .set_units(time="s", distance="m")
        )
        return out

    def test_basic(self, example_patch):
        """Ensure the basic filter works."""
        filt = [2e3, 2.2e3, 8e3, 2e4]
        filtered_patch = example_patch.slope_filter(filt=filt)
        assert isinstance(filtered_patch, dc.Patch)
        assert filtered_patch.shape == example_patch.shape
        assert not np.array_equal(filtered_patch.data, example_patch.data)

    def test_attenuated_slopes(self, event_patch_1):
        """Ensure attenuated slopes are much lower in absolute values."""
        # For some reason when padding isn't performed the attenation can
        # be slightly off, need to look into thos.
        example_patch = event_patch_1.pad(time="fft", distance="fft")
        filt = [2e3, 2.2e3, 8e3, 2e4]
        filtered_patch = example_patch.slope_filter(filt=filt)
        dft_unfiltered = example_patch.dft(("time", "distance"))
        dft_filtered = filtered_patch.dft(("time", "distance"))
        dft_ft_filtered = dft_filtered.slope_filter(filt=filt)
        # Get slope values
        slope = np.abs(self.get_slope_array(dft_filtered))
        # Get values that should be zeroed by filter.
        in_attenuated_range = (slope <= filt[0]) | (slope >= filt[3])
        # Ensure the patch filtered in ft domain is nearly 0
        assert np.allclose(dft_ft_filtered.data[in_attenuated_range], 0)
        assert np.allclose(dft_filtered.data[in_attenuated_range], 0)
        # compare filtered vs unfiltered absolute values
        unfilt = np.abs(dft_unfiltered.data[in_attenuated_range].flatten())
        filt = np.abs(dft_filtered.data[in_attenuated_range].flatten())
        assert np.all(filt < unfilt)

    def test_directional_filter(self, example_patch):
        """Ensure directional logic runs."""
        filtered_patch = example_patch.slope_filter(
            filt=[2e3, 2.2e3, 8e3, 2e4], directional=True
        )

        assert isinstance(filtered_patch, dc.Patch)

    def test_notch_filter(self, example_patch):
        """Ensure notching can be performed with slope filter."""
        filtered_patch = example_patch.slope_filter(
            filt=[2e3, 2.2e3, 8e3, 2e4], notch=True
        )
        assert isinstance(filtered_patch, dc.Patch)

    def test_different_params_not_equal(self, example_patch):
        """Ensure filter is sensitive to different parameters."""
        filtered_patch1 = example_patch.slope_filter(filt=[1e3, 1.5e3, 5e3, 1e4])
        filtered_patch2 = example_patch.slope_filter(filt=[2e3, 2.2e3, 8e3, 2e4])
        assert not np.array_equal(filtered_patch1.data, filtered_patch2.data)

    def test_fk_input_patch(self, example_patch):
        """Ensure an input fk patch is returned in the fk domain."""
        fk_patch = example_patch.dft(dim=("time", "distance"), real="time")
        out = fk_patch.slope_filter(filt=[1e3, 1.5e3, 5e3, 1e4])
        # Ensure the output is still in the fk domain
        dims = set(out.dims)
        assert "ft_time" in dims
        assert "ft_distance" in dims

    def test_other_dims(self, example_patch):
        """Ensure dims other than time and distance work."""
        patch = example_patch.rename_coords(time="money")
        out = patch.slope_filter(
            filt=[2e3, 2.2e3, 8e3, 2e4],
            dims=("money", "distance"),
        )
        assert out.dims == patch.dims

    def test_bad_filt_raises(self, example_patch):
        """Bad filter params should raise Parameter error."""
        msg = "filt must be a sorted length 4 sequence"
        filt = [1e3, 1.5e3, 5e3]
        with pytest.raises(ParameterError, match=msg):
            example_patch.slope_filter(filt=filt)

    def test_bad_dims(self, example_patch):
        """Ensure passing dimensions not in patch raises."""
        patch = example_patch.rename_coords(time="money")
        msg = "are missing from patch"
        filt = [1e3, 1.5e3, 5e3, 10e3]
        with pytest.raises(ParameterError, match=msg):
            patch.slope_filter(filt=filt, dims=("time", "distance"))

    def test_units_raise_no_unit_coords(self, example_patch):
        """Ensure A UnitError is raised if one of the coords doesn't have units."""
        patch = example_patch.set_units(distance="")
        filt = np.array([1e3, 1.5e3, 5e3, 10e3]) * get_unit("m/s")
        with pytest.raises(UnitError):
            patch.slope_filter(filt=filt)

    def test_units(self, example_patch):
        """Ensure units can be specified on filt."""
        filt1 = np.array([1e3, 1.5e3, 5e3, 10e3])
        filt2 = convert_units(filt1, "ft/s", "m/s")
        # All these should provide the same filter.
        out1 = example_patch.slope_filter(filt=filt1)
        out2 = example_patch.slope_filter(filt=filt1 * get_unit("m / s"))
        out3 = example_patch.slope_filter(filt=filt2 * get_unit("ft / s"))

        assert np.allclose(out1.data, out2.data)
        assert np.allclose(out2.data, out3.data)

    def test_inverted_units(self, example_patch):
        """Ensure units are automatically inverted (eg slowness should work)"""
        filt = np.array([1e3, 1.5e3, 5e3, 10e3])
        slowness = np.sort(1 / filt * get_unit("s/m"))
        out1 = example_patch.slope_filter(filt=slowness)
        out2 = example_patch.slope_filter(filt=filt * get_unit("m/s"))
        assert np.allclose(out1.data, out2.data)

    def test_units_list(self, example_patch):
        """Ensure units as a list still work (see #463)."""
        speed = 5_000 * dc.get_quantity("m/s")
        filt = [speed * 0.90, speed * 0.95, speed * 1.05, speed * 1.1]
        # The test passes if this line doesn't raise an error.
        out = example_patch.slope_filter(filt)
        assert isinstance(out, dc.Patch)
